# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import datetime
import logging
import threading
import typing

import requests

import ccc.concourse
import ccc.elasticsearch
import ccc.github
import ccc.secrets_server
import ci.util
import concourse.client.api
import concourse.client.model
import concourse.enumerator
import concourse.replicator
import model
import whd.model
import whd.pull_request
import whd.util

from github3.exceptions import NotFoundError

from .pipelines import replicate_repository_pipelines
from concourse.client.util import determine_jobs_to_be_triggered
from concourse.enumerator import JobMappingNotFoundError
from concourse.model.job import AbortObsoleteJobs
from model import ConfigFactory
from model.base import ConfigElementNotFoundError
from model.webhook_dispatcher import WebhookDispatcherConfig

from concourse.client.model import (
    ResourceType,
)
from .model import (
    AbortConfig,
    Pipeline,
    PullRequestAction,
    PullRequestEvent,
    PushEvent,
    RefType,
)
import whd.metric


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class GithubWebhookDispatcher:
    def __init__(
        self,
        cfg_factory,
        cfg_set,
        whd_cfg: WebhookDispatcherConfig
    ):
        self.cfg_factory: model.ConfigFactory = cfg_factory
        self.cfg_set = cfg_set
        self.whd_cfg = whd_cfg
        logger.info(f'github-whd initialised for cfg-set: {cfg_set.name()}')

    def concourse_clients(
        self,
    ) -> typing.Generator[concourse.client.api.ConcourseApiBase, None, None]:
        for concourse_config_name in self.whd_cfg.concourse_config_names():
            concourse_cfg = self.cfg_factory.concourse(concourse_config_name)
            job_mapping_set = self.cfg_factory.job_mapping(concourse_cfg.job_mapping_cfg_name())
            for job_mapping in job_mapping_set.job_mappings().values():
                yield ccc.concourse.client_from_cfg_name(
                    concourse_cfg_name=concourse_cfg.name(),
                    team_name=job_mapping.team_name(),
                )

    def dispatch_create_event(
        self,
        create_event,
        delivery_id: str,
        repository: str,
        hostname: str,
        es_client: ccc.elasticsearch.ElasticSearchClient,
        dispatch_start_time: datetime.datetime,
    ):
        ref_type = create_event.ref_type()
        if not ref_type == RefType.BRANCH:
            logger.info(f'ignored create event with type {ref_type}')
            return

        thread = threading.Thread(
            target=self._update_pipeline_definition,
            kwargs={
                'event': create_event,
                'delivery_id': delivery_id,
                'repository': repository,
                'hostname': hostname,
                'es_client': es_client,
                'dispatch_start_time': dispatch_start_time,
            }
        )
        thread.start()

    def dispatch_push_event(
        self,
        push_event,
        delivery_id: str,
        repository: str,
        hostname: str,
        es_client: ccc.elasticsearch.ElasticSearchClient,
        dispatch_start_time: datetime.datetime,
    ):
        def process_push_event(
            delivery_id,
            hostname,
            es_client,
            repository,
            dispatch_start_time,
            event,
            event_type,
        ):
            if self._pipeline_definition_changed(event):
                try:
                    self._update_pipeline_definition(
                        event=event,
                        delivery_id=delivery_id,
                        repository=repository,
                        hostname=hostname,
                        es_client=es_client,
                        dispatch_start_time=dispatch_start_time,
                    )
                except ValueError as e:
                    logger.warning(
                        f'Received error updating pipeline-definitions: "{e}". '
                        'Will still abort running jobs (if configured) and trigger resource checks.'
                    )

            self.abort_running_jobs_if_configured(event)

            for concourse_api in self.concourse_clients():
                logger.debug(f'using concourse-api: {concourse_api}')
                resources = self._matching_resources(
                    concourse_api=concourse_api,
                    event=event,
                )
                logger.debug('triggering resource-check')
                whd.util.trigger_resource_check(concourse_api=concourse_api, resources=resources)

            process_end_time = datetime.datetime.now()
            process_total_seconds = (
                process_end_time - dispatch_start_time
            ).total_seconds()
            webhook_delivery_metric = whd.metric.WebhookDelivery.create(
                delivery_id=delivery_id,
                event_type=event_type,
                repository=repository,
                hostname=hostname,
                process_total_seconds=process_total_seconds,
            )
            if not es_client:
                return

            ccc.elasticsearch.metric_to_es(
                es_client=es_client,
                metric=webhook_delivery_metric,
                index_name=whd.metric.index_name(webhook_delivery_metric),
            )

        thread = threading.Thread(
            target=process_push_event,
            kwargs={
                'delivery_id': delivery_id,
                'hostname': hostname,
                'es_client': es_client,
                'repository': repository,
                'event_type': 'push',
                'dispatch_start_time': dispatch_start_time,
                'event': push_event,
            }
        )
        thread.start()

    def _update_pipeline_definition(
        self,
        event,
        delivery_id: str,
        repository: str,
        hostname: str,
        es_client: ccc.elasticsearch.ElasticSearchClient,
        dispatch_start_time: datetime.datetime,
    ):
        def _do_update(
            delivery_id: str,
            event_type: str,
            repository: str,
            hostname: str,
            dispatch_start_time: datetime.datetime,
            es_client: ccc.elasticsearch.ElasticSearchClient,
        ):
            repo_url = event.repository().repository_url()
            job_mapping_set = self.cfg_set.job_mapping()
            job_mapping = job_mapping_set.job_mapping_for_repo_url(repo_url, self.cfg_set)

            replicate_repository_pipelines(
                repo_url=repo_url,
                cfg_set=self.cfg_factory.cfg_set(job_mapping.replication_ctx_cfg_set()),
                whd_cfg=self.whd_cfg,
            )

            process_end_time = datetime.datetime.now()
            process_total_seconds = (process_end_time - dispatch_start_time).total_seconds()
            webhook_delivery_metric = whd.metric.WebhookDelivery.create(
                delivery_id=delivery_id,
                event_type=event_type,
                repository=repository,
                hostname=hostname,
                process_total_seconds=process_total_seconds,
            )
            if es_client:
                ccc.elasticsearch.metric_to_es(
                    es_client=es_client,
                    metric=webhook_delivery_metric,
                    index_name=whd.metric.index_name(webhook_delivery_metric),
                )

        try:
            _do_update(
                delivery_id=delivery_id,
                event_type='create',
                repository=repository,
                hostname=hostname,
                dispatch_start_time=dispatch_start_time,
                es_client=es_client,
            )
        except (JobMappingNotFoundError, ConfigElementNotFoundError) as e:
            # A config element was missing or o JobMapping for the given repository was present.
            # Print warning, reload and try again
            logger.warning(
                f'failed to update pipeline definition: {e}. Will reload config and try again.'
            )
            # Attempt to fetch latest cfg from SS and replace it
            raw_dict = ccc.secrets_server.SecretsServerClient.default().retrieve_secrets()
            self.cfg_factory = ConfigFactory.from_dict(raw_dict)
            self.cfg_set = self.cfg_factory.cfg_set(self.cfg_set.name())
            # retry
            _do_update(
                delivery_id=delivery_id,
                event_type='create',
                repository=repository,
                hostname=hostname,
                dispatch_start_time=dispatch_start_time,
                es_client=es_client,
            )

    def _pipeline_definition_changed(self, push_event):
        if '.ci/pipeline_definitions' in push_event.modified_paths():
            return True
        return False

    def determine_affected_pipelines(self, push_event) -> typing.Generator[Pipeline, None, None]:
        '''yield each concourse pipeline that may be affected by the given push-event.
        '''
        repo = push_event.repository()
        repo_url = repo.repository_url()
        job_mapping_set = self.cfg_set.job_mapping()

        try:
            job_mapping = job_mapping_set.job_mapping_for_repo_url(repo_url, self.cfg_set)
        except ValueError:
            logger.info(f'no job-mapping found for {repo_url=} - will not interact w/ pipeline(s)')
            return

        try:
            repo_enumerator = concourse.enumerator.GithubRepositoryDefinitionEnumerator(
                repository_url=repo_url,
                cfg_set=self.cfg_factory.cfg_set(job_mapping.replication_ctx_cfg_set()),
            )
        except concourse.enumerator.JobMappingNotFoundError:
            logger.info(f'no job-mapping matched for {repo_url=} - will not interact w/ pipeline(s)')
            return

        try:
            definition_descriptors = [d for d in repo_enumerator.enumerate_definition_descriptors()]
        except NotFoundError:
            logger.warning(
                f"Unable to access repository '{repo_url}' on github '{repo.github_host()}'. "
                "Please make sure the repository exists and the technical user has the necessary "
                "permissions to access it."
            )
            definition_descriptors = []

        for descriptor in definition_descriptors:
            # need to merge and consider the effective definition
            effective_definition = descriptor.pipeline_definition
            for override in descriptor.override_definitions:
                effective_definition = ci.util.merge_dicts(effective_definition, override)

            yield Pipeline(
                pipeline_name=descriptor.effective_pipeline_name(),
                target_team=descriptor.concourse_target_team,
                effective_definition=effective_definition,
            )

    def matching_client(self, team):
        for c in self.concourse_clients():
            if c.routes.team == team:
                return c

    def abort_running_jobs_if_configured(self, push_event):
        builds_to_consider = 5
        for pipeline in self.determine_affected_pipelines(
            push_event
        ):
            if not (client := self.matching_client(pipeline.target_team)):
                logger.info(
                    f'no matching job-mapping for {pipeline.pipeline_name=} - skipping abortion'
                )
                continue

            try:
                pipeline_config = client.pipeline_cfg(pipeline.pipeline_name)
            except requests.exceptions.HTTPError as e:
                # might not exist yet if the pipeline was just rendered by the WHD
                if e.response.status_code != 404:
                    raise e
                logger.warning(f"could not retrieve pipeline config for '{pipeline.pipeline_name}'")
                continue

            resources = [
                r for r in pipeline_config.resources
                if ResourceType(r.type) in (ResourceType.GIT, ResourceType.PULL_REQUEST)
            ]
            for job in determine_jobs_to_be_triggered(*resources):
                if (
                    not pipeline.effective_definition['jobs'].get(job.name)
                    or not 'abort_outdated_jobs' in pipeline.effective_definition['jobs'][job.name]
                ):
                    continue
                abort_cfg = AbortConfig.from_dict(
                    pipeline.effective_definition['jobs'][job.name]
                )

                if abort_cfg.abort_obsolete_jobs is AbortObsoleteJobs.NEVER:
                    continue
                elif (
                    abort_cfg.abort_obsolete_jobs is AbortObsoleteJobs.ON_FORCE_PUSH_ONLY
                    and not push_event.is_forced_push()
                ):
                    continue
                elif abort_cfg.abort_obsolete_jobs is AbortObsoleteJobs.ALWAYS:
                    pass
                else:
                    raise NotImplementedError(abort_cfg.abort_obsolete_jobs)

                running_builds = [
                    b for b in client.job_builds(pipeline.pipeline_name, job.name)
                    if b.status() is concourse.client.model.BuildStatus.RUNNING
                ][:builds_to_consider]

                for build in running_builds:
                    if build.plan().contains_version_ref(push_event.previous_ref()):
                        logger.info(
                            f"Aborting obsolete build '{build.build_number()}' for job '{job.name}'"
                        )
                        client.abort_build(build.id())

    def dispatch_pullrequest_event(
        self,
        pr_event: whd.model.PullRequestEvent,
        es_client: ccc.elasticsearch.ElasticSearchClient,
        dispatch_start_time: datetime.datetime,
    ) -> bool:
        '''Process the given push event.

        Return `True` if event will be processed, `False` if no processing will be done.
        '''
        if not pr_event.action() in (
            PullRequestAction.OPENED,
            PullRequestAction.REOPENED,
            PullRequestAction.LABELED,
            PullRequestAction.SYNCHRONIZE,
        ):
            logger.info(f'ignoring pull-request action {pr_event.action()}')
            return False

        thread = threading.Thread(
            target=whd.pull_request.process_pr_event,
            kwargs={
                'concourse_clients': self.concourse_clients(),
                'cfg_factory': self.cfg_factory,
                'whd_cfg': self.whd_cfg,
                'cfg_set': self.cfg_set,
                'pr_event': pr_event,
                'es_client': es_client,
                'dispatch_start_time': dispatch_start_time,
            }
        )
        thread.start()

        return True

    def _matching_resources(
        self,
        concourse_api: concourse.client.api.ConcourseApiBase,
        event,
    ) -> typing.Generator[concourse.client.model.PipelineConfigResource, None, None]:
        if isinstance(event, PushEvent):
            resource_type = ResourceType.GIT
        elif isinstance(event, PullRequestEvent):
            resource_type = ResourceType.PULL_REQUEST
        else:
            raise NotImplementedError

        resources_gen = concourse_api.pipeline_resources(
            concourse_api.pipelines(),
            resource_type=resource_type,
        )

        for resource in resources_gen:
            resource: concourse.client.model.PipelineConfigResource

            ghs = resource.github_source()
            repository = event.repository()
            if not ghs.hostname() == repository.github_host():
                continue
            if not ghs.repo_path().lstrip('/') == repository.repository_path():
                continue
            if isinstance(event, PushEvent):
                if not event.ref().endswith(ghs.branch_name()):
                    continue
                if msg := event.commit_message():
                    if (
                        not ghs.disable_ci_skip()
                        and any(skip in msg for skip in ('[skip ci]', '[ci skip]'))
                    ):
                        logger.info(
                            f"Do not trigger resource {resource.name}. Found [skip ci] or [ci skip]"
                        )
                        continue

            yield resource



import datetime
import dataclasses
from enum import Enum, IntEnum
from functools import total_ordering
import typing

import github.compliance.model as gcm


class ScanStatusValues(Enum):
    NEW = 1
    PRE_SCAN = 2
    QUEUED = 3
    SCANNING = 4
    POST_SCAN = 6
    FINISHED = 7
    CANCELED = 8
    FAILED = 9
    SOURCE_PULLING_AND_DEPLOYMENT = 10


class CustomFieldKeys(Enum):
    COMPONENT_NAME = 4
    VERSION = 5
    HASH = 6


@dataclasses.dataclass
class CustomField:
    id: int
    value: str


@total_ordering
class Severity(IntEnum):
    INFO = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3

    @classmethod
    def from_str(cls, s: str) -> IntEnum | None:
        su = s.upper()
        if su == 'INFO':
            return Severity.INFO
        elif su == 'LOW':
            return Severity.LOW
        elif su == 'MEDIUM':
            return Severity.MEDIUM
        elif su == 'HIGH':
            return Severity.HIGH
        else:
            return None

    def __str__(self):
        return self.name.lower()


@dataclasses.dataclass
class ProjectDetails:
    id: int
    teamId: int
    name: str
    customFields: typing.List[CustomField] = dataclasses.field(default_factory=list)

    def get_custom_field(self, attribute_key: CustomFieldKeys, pop: bool = False):
        if not isinstance(attribute_key, CustomFieldKeys):
            raise ValueError(attribute_key)

        for cf in self.customFields:
            if cf.id == attribute_key.value:
                if pop:
                    self.customFields.remove(cf)
                return cf.value

    def set_custom_field(self, attribute_key: CustomFieldKeys, value: str):
        self.get_custom_field(attribute_key=attribute_key, pop=True)
        self.customFields.append(CustomField(id=attribute_key.value, value=value))


@dataclasses.dataclass
class ScanStatusDetails:
    stage: str
    step: str


@dataclasses.dataclass
class ScanStatus:
    id: int
    name: str
    details: ScanStatusDetails


@dataclasses.dataclass
class AuthResponse:
    access_token: str
    expires_in: int
    token_type: str
    expires_at: datetime.datetime = None

    def is_valid(self):
        return datetime.datetime.now() < self.expires_at


@dataclasses.dataclass
class ScanDateAndTime:
    startedOn: str
    finishedOn: typing.Optional[str]


@dataclasses.dataclass
class ScanResponse:
    owner: str
    id: int
    scanRisk: int
    scanRiskSeverity: int
    status: ScanStatus
    isIncremental: bool
    owningTeamId: int
    dateAndTime: typing.Optional[ScanDateAndTime] = None

    def status_value(self):
        return ScanStatusValues(self.status.id)


@dataclasses.dataclass
class ScanStatistic:
    highSeverity: int
    mediumSeverity: int
    lowSeverity: int
    infoSeverity: int
    statisticsCalculationDate: str


@dataclasses.dataclass
class ScanSettings:
    projectId: int
    isIncremental: bool = True
    isPublic: bool = True
    forceScan: bool = True
    comment: str = ""


# below types are not used for http body deserialization
@dataclasses.dataclass
class ScanResult(gcm.ScanResult):
    """
    ScanResult is a data container for a scan result for a component version.
    Successful scans must provide not None values for `scan_response`, `scan_statistic`, and
    `report_url`.
    """
    artifact_name: str
    project_id: int
    scan_response: ScanResponse | None
    scan_statistic: ScanStatistic | None
    report_url: str | None
    overview_url: str


@dataclasses.dataclass
class FinishedScans:
    failed_scans: typing.List[ScanResult] = dataclasses.field(default_factory=list)
    scans: typing.List[ScanResult] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class FailedScan:
    artifact_name: str



import logging
import typing

import boto3

import ci.util
import ctx
import model.aws

logger = logging.getLogger(__name__)


def session(aws_cfg: typing.Union[str, model.aws.AwsProfile], region_name=None):
    if isinstance(aws_cfg, str):
        cfg_factory = ci.util.ctx().cfg_factory()
        aws_cfg = cfg_factory.aws(aws_cfg)

    region_name = region_name or aws_cfg.region()

    session = boto3.Session(
        aws_access_key_id=aws_cfg.access_key_id(),
        aws_secret_access_key=aws_cfg.secret_access_key(),
        region_name=region_name,
    )

    return session


def default_session(
    cfg_factory=None,
    cfg_set=None,
):
    if not cfg_set and not cfg_factory:
        cfg_factory = ctx.cfg_factory()

    if not cfg_set:
        cfg_set = cfg_factory.cfg_set(ci.util.current_config_set_name())

    try:
        cfg = cfg_set.aws()
        return session(cfg)
    except:
        logger.warning('failed to retrieve default aws cfg')
        return None



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import abc
import enum
import reutil

import ocm


class FilterBase:
    @abc.abstractmethod
    def matches(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
    ):
        pass


class MatchAllFilter(FilterBase):
    '''
    A filter matching everything. Use at the end of a processing.cfg
    '''
    def matches(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
    ):
        return True


class ImageFilter(FilterBase):
    def __init__(
        self,
        include_image_refs=(),
        exclude_image_refs=(),
        include_image_names=(),
        exclude_image_names=(),
        include_artefact_types=(),
        exclude_artefact_types=(),
    ):
        self._image_ref_filter = reutil.re_filter(
            include_regexes=include_image_refs,
            exclude_regexes=exclude_image_refs,
            value_transformation=lambda oci_resource: oci_resource.access.imageReference,
        )
        self._image_name_filter = reutil.re_filter(
            include_regexes=include_image_names,
            exclude_regexes=exclude_image_names,
            value_transformation=lambda resource: resource.name,
        )
        self._include_artefact_types = include_artefact_types
        self._exclude_artefact_types = exclude_artefact_types

    def matches(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
    ):
        if resource.access.type is not ocm.AccessType.OCI_REGISTRY:
            return False

        name_matches = self._image_ref_filter(resource) and \
            self._image_name_filter(resource)

        if not name_matches:
            return False

        if isinstance(resource.type, enum.Enum):
            type_name = resource.type.value
        elif isinstance(resource.type, str):
            type_name = resource.type
        elif resource.type is None:
            return name_matches
        else:
            raise ValueError(resource.type)

        for exclude_type_name in self._exclude_artefact_types:
            try:
                exclude_type = ocm.ArtefactType(exclude_type_name)
                if exclude_type is resource.type:
                    return False # type was explicitly excluded
            except ValueError:
                # fallback to str-comparison
                if exclude_type_name == type_name:
                    return False

        if not self._include_artefact_types:
            return name_matches # if no types are explicitly configured to be included, and
            # this line is reached (which means the type was also not explicitly excluded),
            # then match depending on name

        for include_type_name in self._include_artefact_types:
            try:
                include_type = ocm.ArtefactType(include_type_name)
                if include_type is resource.type:
                    return True
            except ValueError:
                # fallback to str-comparison
                if include_type_name == type_name:
                    return True
        else:
            # if types for inclusion are explicitly passed, and this line is reached, it means
            # there was not match
            return False


class ComponentFilter(FilterBase):
    def __init__(
        self,
        include_component_names=(),
        exclude_component_names=(),
    ):
        self._comp_name_filter = reutil.re_filter(
            include_regexes=include_component_names,
            exclude_regexes=exclude_component_names,
            value_transformation=lambda component: component.name,
        )

    def matches(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
    ):
        return self._comp_name_filter(component)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import functools
import json
import logging
import os

import concourse.client.model
import concourse.model.traits.meta
import concourse.steps.meta

from concourse.client.routes import ConcourseApiRoutesBase
from ci.util import (
    _running_on_ci,
    check_env,
    ctx,
)
import ccc.concourse
import ccc.github
import ci.log

logger = logging.getLogger()
ci.log.configure_default_logging()


@dataclasses.dataclass
class PipelineMetaData:
    pipeline_name: str
    job_name: str
    current_config_set_name: str
    team_name: str


def get_pipeline_metadata():
    if not _running_on_ci():
        raise RuntimeError('Pipeline-metadata is only available if running on CI infrastructure')

    current_cfg_set_name = check_env('CONCOURSE_CURRENT_CFG')
    team_name = check_env('CONCOURSE_CURRENT_TEAM')
    pipeline_name = check_env('PIPELINE_NAME')
    job_name = check_env('BUILD_JOB_NAME')

    return PipelineMetaData(
        pipeline_name=pipeline_name,
        job_name=job_name,
        current_config_set_name=current_cfg_set_name,
        team_name=team_name,
    )


def _current_concourse_config(cfg_factory=None):
    if not _running_on_ci():
        raise RuntimeError('Can only determine own concourse config if running on CI')

    if cfg_factory:
        cfg_set = cfg_factory.cfg_set(check_env('CONCOURSE_CURRENT_CFG'))
        concourse_cfg = cfg_set.concourse()
    else:
        concourse_cfg = ctx().cfg_set().concourse()

    return concourse_cfg


def own_running_build_url(cfg_factory=None) -> str:
    if not _running_on_ci():
        raise RuntimeError('Can only determine own build url if running on CI infrastructure')

    pipeline_metadata = get_pipeline_metadata()

    own_build = find_own_running_build(cfg_factory=cfg_factory)
    cc_cfg = _current_concourse_config(cfg_factory=cfg_factory)

    return ConcourseApiRoutesBase.running_build_url(
        cc_cfg.external_url(),
        pipeline_metadata,
        own_build.build_number(),
    )


def meta_info_file_from_env() -> str:
    return os.path.abspath(
        os.path.join(
            check_env('CC_ROOT_DIR'),
            concourse.model.traits.meta.DIR_NAME,
            concourse.steps.meta.jobmetadata_filename,
        )
    )


def has_metadata() -> bool:
    return os.path.isfile(
        meta_info_file_from_env()
    )


@functools.lru_cache()
def find_own_running_build(cfg_factory=None) -> concourse.client.model.Build:
    '''
    Determines the current build job running on concourse by relying on the "meta" contract (
    see steps/meta), which prints a JSON document containing a UUID. By iterating through all
    current build jobs (considering running jobs only), and comparing the UUID read via file
    system and the UUID from build log output, it is possible to tell whether or not a given
    build job is the one from which this function was invoked.
    '''
    if not _running_on_ci():
        raise RuntimeError('Can only find own running build if running on CI infrastructure.')

    meta_info_file = meta_info_file_from_env()

    with open(meta_info_file, 'r') as f:
        metadata_json = json.load(f)

    build_job_uuid = metadata_json['uuid']

    if cfg_factory:
        current_cfg_set_name = check_env('CONCOURSE_CURRENT_CFG')
        cfg_set = cfg_factory.cfg_set(current_cfg_set_name)
    else:
        cfg_set = None

    pipeline_metadata = get_pipeline_metadata()
    client = ccc.concourse.client_from_env(
        team_name=pipeline_metadata.team_name,
        cfg_set=cfg_set,
    )

    # only consider limited amount of jobs to avoid large number of requests in case we do not
    # find ourself (assumption: there are only few running jobs in parallel at a given time)
    consider_builds = 20
    builds = client.job_builds(pipeline_metadata.pipeline_name, pipeline_metadata.job_name)
    builds = [
        build for build in builds
        if build.status() is concourse.client.model.BuildStatus.RUNNING
    ][:consider_builds]

    # avoid parsing too much output. usually, there will be only one line (our JSON output)
    # sometimes (new image version is retrieved), there will be a few lines more.
    for build in builds:
        build_events = build.events()
        build_plan = build.plan()
        meta_task_id = build_plan.task_id(concourse.model.traits.meta.META_STEP_NAME)

        # we expect output to only contain valid JSON
        meta_output = ''.join(build_events.iter_buildlog(meta_task_id)).strip()

        try:
            uuid_json = json.loads(meta_output)
        except json.decoder.JSONDecodeError:
            logger.error(f'Error when parsing {meta_output=}')
            continue # ignore - we might still find "our" job
        if uuid_json['uuid'] == build_job_uuid:
            return build
    else:
        raise RuntimeError('Could not determine own Concourse job.')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import oci.model as om

import ctt.oci_platform as platform


def test_platform_filter():
    filter_func = platform.PlatformFilter(
        included_platform_regexes=(
            'linux/.*',
        ),
    )

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='amd64',
    ))
    assert matches is True

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='arm64',
        variant='v6',
    ))
    assert matches is True

    matches = filter_func(om.OciPlatform(
        os='darwin',
        architecture='arm64',
    ))
    assert matches is False

    filter_func = platform.PlatformFilter(
        included_platform_regexes=(
            'linux/arm64',
        ),
    )

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='arm64',
    ))
    assert matches is True

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='arm64',
        variant='v7',
    ))
    assert matches is False

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='amd64',
    ))
    assert matches is False

    filter_func = platform.PlatformFilter(
        included_platform_regexes=(
            'linux/arm64/v7',
        ),
    )

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='arm64',
        variant='v7',
    ))
    assert matches is True

    matches = filter_func(om.OciPlatform(
        os='linux',
        architecture='arm64',
        variant='v6',
    ))
    assert matches is False

    filter_func = platform.PlatformFilter(
        included_platform_regexes=(
            '.*/arm64/.*',
        ),
    )

    matches = filter_func(om.OciPlatform(
        os='ios',
        architecture='arm64',
        variant='v7',
    ))
    assert matches is True

    matches = filter_func(om.OciPlatform(
        os='ios',
        architecture='arm',
        variant='v7',
    ))
    assert matches is False



import base64
import copy
import logging
import time
import typing

import kubernetes.client
import kubernetes.config

import cfg_mgmt
import cfg_mgmt.util
import ci.util
import model
import model.kubernetes

from cfg_mgmt.model import (
    CfgQueueEntry,
    ValidationError,
)
from kubernetes.client import (
    CoreV1Api,
    V1ObjectMeta,
    V1Secret,
    V1ServiceAccount,
)


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


def rotate_cfg_element(
    cfg_element: model.kubernetes.KubernetesConfig,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:

    # TODO: Cannot rotate kubeconfig without SA

    # copy passed cfg_element, since we update in-place.
    raw_cfg = copy.deepcopy(cfg_element.raw)
    cfg_to_rotate = model.kubernetes.KubernetesConfig(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )

    api_client = kubernetes.config.new_client_from_config_dict(cfg_element.kubeconfig())
    core_api = CoreV1Api(api_client)

    # find service-account
    service_account_config = cfg_to_rotate.service_account()

    sa: V1ServiceAccount = core_api.read_namespaced_service_account(
        name=service_account_config.name(),
        namespace=service_account_config.namespace(),
    )

    if cfg_element.rotation_strategy() == model.kubernetes.RotationStrategy.TOKEN_REQUEST:
        if not (bound_secret_name := cfg_to_rotate.raw.get('bound_secret_name')):
            raise ValueError(
                'Token request rotation requires attribute "bound_secret_name" in kubernetes-cfg.'
            )

        secret_id = {
            'old_tokens': [{"name": bound_secret_name}],
        }

        binding_secret_name = _create_secret_for_token_bind(
            core_api=core_api,
            service_account=sa,
        )

        expiration_seconds = 60 * 60 * 24 * 90 # 7776000s = 90d is max
        new_sa_access_token = core_api.create_namespaced_service_account_token(
            name=service_account_config.name(),
            namespace=service_account_config.namespace(),
            body={
                "spec": {"expirationSeconds": expiration_seconds},
                "bound_object_ref": {"name": binding_secret_name, "kind": "Secret"},
            },
        )

        if not new_sa_access_token:
            # the token was not created for service account. Clean up
            # and raise
            core_api.delete_namespaced_secret(
                name=binding_secret_name,
                namespace=service_account_config.namespace(),
            )
            raise RuntimeError('Error creating new service account token.')

        # update kubeconfig with new token
        cfg_to_rotate.kubeconfig()["users"][0]["user"]["token"] = new_sa_access_token.status.token

        cfg_to_rotate.raw["bound_secret_name"] = binding_secret_name

        def revert():
            logger.warning(
                f"An error occurred during update of kubernetes config '{cfg_to_rotate.name()}', "
                'rolling back'
            )

            # delete newly created service-account token
            core_api.delete_namespaced_secret(
                name=binding_secret_name,
                namespace=service_account_config.namespace(),
            )

    elif cfg_element.rotation_strategy() == model.kubernetes.RotationStrategy.SECRET:
        old_tokens = [{"name": token.name} for token in sa.secrets] if sa.secrets else []
        secret_id = {
            'old_tokens': old_tokens,
        }

        # create new token
        new_access_token_name = _create_token_for_sa(
            core_api=core_api,
            service_account=sa,
        )

        new_access_token = _wait_for_token_secret(
            core_api=core_api,
            token_name=new_access_token_name,
            token_namespace=service_account_config.namespace(),
        )

        if not new_access_token:
            # the cluster controller did not update the secret in the given time. Clean up
            # and raise
            core_api.delete_namespaced_secret(
                new_access_token_name,
                service_account_config.namespace(),
            )
            raise RuntimeError(
                'Service account token-secret was not populated with the required data in time.'
            )

        # patch service-account with new secret (this updates the secret that will be mounted to
        # k8s-pods)
        patched_sa = _update_sa(
            core_api=core_api,
            service_account=sa,
            token_list=[{'name': new_access_token_name}]
        )

        # create new kubeconfig from new access token
        updated_kubeconfig = _update_kubeconfig(
            kubeconfig=cfg_element.kubeconfig(),
            patched_service_account=patched_sa,
            new_access_token=new_access_token,
        )
        cfg_to_rotate.raw['kubeconfig'] = updated_kubeconfig

        def revert():
            logger.warning(
                f"An error occurred during update of kubernetes config '{cfg_to_rotate.name()}', "
                'rolling back'
            )
            # revert changes to service-account
            _update_sa(
                core_api=core_api,
                service_account=patched_sa,
                token_list=old_tokens,
            )

            # delete newly created service-account token
            _delete_sa_token(
                core_api=core_api,
                namespace=service_account_config.namespace(),
                token_name=new_access_token_name,
            )

    else:
        raise NotImplementedError(f'{cfg_element.rotation_strategy()=}')

    return revert, secret_id, cfg_to_rotate


def _update_sa(
    core_api: CoreV1Api,
    service_account: V1ServiceAccount,
    token_list: typing.List[dict],
) -> V1ServiceAccount:
    service_account_name = service_account.metadata.name
    service_account_namespace = service_account.metadata.namespace
    body = [{
        "op": "replace",
        "path": "/secrets",
        "value": token_list,
    }]
    return core_api.patch_namespaced_service_account(
        name=service_account_name,
        namespace=service_account_namespace,
        body=body,
    )


def _create_token_for_sa(
    core_api: CoreV1Api,
    service_account: V1ServiceAccount,
) -> str:
    service_account_name = service_account.metadata.name
    service_account_namespace = service_account.metadata.namespace
    token = core_api.create_namespaced_secret(
        namespace=service_account_namespace,
        body=V1Secret(
            api_version='v1',
            kind='Secret',
            metadata=V1ObjectMeta(
                generate_name=f'{service_account_name}-token-',
                annotations={'kubernetes.io/service-account.name': service_account_name},
            ),
            type='kubernetes.io/service-account-token',
        ),
    )
    # not all required values are set on the returned object yet. Return only name so that we can
    # fetch it later (name will be generated by the kube-apiserver)
    return token.metadata.name


def _create_secret_for_token_bind(
    core_api: CoreV1Api,
    service_account: V1ServiceAccount,
) -> str:
    service_account_name = service_account.metadata.name
    service_account_namespace = service_account.metadata.namespace
    token = core_api.create_namespaced_secret(
        namespace=service_account_namespace,
        body=V1Secret(
            api_version='v1',
            kind='Secret',
            metadata=V1ObjectMeta(
                generate_name=f'{service_account_name}-token-bind-',
            ),
            type='Opaque',
        ),
    )
    # not all required values are set on the returned object yet. Return only name so that we can
    # store it for deleting on next rotation
    return token.metadata.name


def _delete_sa_token(
    core_api: CoreV1Api,
    token_name: str,
    namespace: str,
):
    core_api.delete_namespaced_secret(
        name=token_name,
        namespace=namespace,
    )


def _update_kubeconfig(
    kubeconfig: dict,
    patched_service_account: V1ServiceAccount,
    new_access_token: V1Secret,
) -> dict:
    kubeconfig = copy.deepcopy(kubeconfig)

    sa_name = patched_service_account.metadata.name
    namespace = patched_service_account.metadata.namespace

    crt = new_access_token.data.get('ca.crt')
    token = new_access_token.data.get('token')
    kubeconfig['clusters'][0]['cluster']['certificate-authority-data'] = crt
    kubeconfig['users'] = [{
        'user': {'token': base64.b64decode(token).decode("utf-8")},
        'name': kubeconfig['contexts'][0]['context']['user']
    }]
    # Note: This is only our convention. We are free to choose whatever we want as far
    # as k8s is concerned (context-name is referred to as "nickname") as long as we use
    # it consistently.
    kubeconfig['contexts'][0]['name'] = f'system:serviceaccount:{namespace}:{sa_name}'
    kubeconfig['contexts'][0]['context']['namespace'] = namespace
    kubeconfig['current-context'] = f'system:serviceaccount:{namespace}:{sa_name}'

    return kubeconfig


def _wait_for_token_secret(
    core_api: CoreV1Api,
    token_name: str,
    token_namespace: str,
    retries: int = 3,
    interval: int = 5,
) -> typing.Union[V1Secret, None]:
    new_access_token = core_api.read_namespaced_secret(
        name=token_name,
        namespace=token_namespace,
    )
    if not new_access_token.data:
        logger.info('Access token was not yet populated with required information.')
        if retries > 0:
            logger.info(f'Will sleep for {interval} seconds and try again {retries} more times.')
            time.sleep(interval)
            return _wait_for_token_secret(
                core_api=core_api,
                token_name=token_name,
                token_namespace=token_namespace,
                retries=retries-1,
                interval=interval,
            )
        else:
            return None
    return new_access_token


def delete_config_secret(
    cfg_element: model.kubernetes.KubernetesConfig,
    cfg_factory: model.ConfigFactory,
    cfg_queue_entry: CfgQueueEntry,
) -> model.kubernetes.KubernetesConfig | None:
    api_client = kubernetes.config.new_client_from_config_dict(cfg_element.kubeconfig())
    core_api = CoreV1Api(api_client)

    for token in cfg_queue_entry.secretId['old_tokens']:
        _delete_sa_token(
            core_api=core_api,
            token_name=token['name'],
            namespace=cfg_element.service_account().namespace(),
        )

    return None


def validate_for_rotation(
    cfg_element: model.kubernetes.KubernetesConfig,
):
    if not cfg_element.service_account():
        raise ValidationError("Cannot rotate kubeconfigs without service account configs.")
    if len(cfg_element.kubeconfig()['contexts']) != 1:
        raise ValidationError("Rotation of kubeconfigs with multiple contexts is not implemented.")



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class PostgresCredentials(BasicCredentials):
    def username(self):
        return self.raw.get('username')

    def password(self):
        return self.raw.get('password')


class DeliveryDbConfig(NamedModelElement):
    def credentials(self):
        return PostgresCredentials(self.raw.get('credentials'))

    def hostname(self):
        return self.raw.get('hostname')

    def port(self):
        return self.raw.get('port')

    def helm_values(self):
        return self.raw.get('helm_values')

    def db_type(self):
        return self.raw.get('db_type', 'postgresql')

    def as_url(self):
        creds = self.credentials()
        if self.credentials():
            auth_str = f'{creds.username()}:{creds.passwd()}@'
        else:
            auth_str = ''

        return f'{self.db_type()}://{auth_str}{self.hostname()}:{self.port()}'

    def _defaults_dict(self):
        return {
            'db_type': 'postgresql',
        }

    def _required_attributes(self):
        return (
            'credentials',
        )



import typing

import ci.util
import dacite

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
)
from concourse.model.base import (
    AttributeSpec,
    ModelBase,
    Trait,
    TraitTransformer,
    ScriptType,
)
import concourse.model.traits.component_descriptor

from concourse.model.traits.image_scan import (
    GithubIssueTemplateCfg,
    IssuePolicies,
    Notify,
)


CHECKMARX_ATTRIBUTES = (
    AttributeSpec.required(
        name='team_id',
        doc='checkmarx team id',
        type=int,
    ),
    AttributeSpec.optional(
        name='severity_threshold',
        default='medium',
        doc='threshold for creating issues (high, medium, low, info)',
        type=str,
    ),
    AttributeSpec.required(
        name='cfg_name',
        doc='config name for checkmarx',
        type=str,
    ),
    AttributeSpec.optional(
        name='include_path_regexes',
        doc='paths which should be included in the scan',
        default=(),
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        doc='paths which should be excluded in the scan',
        default=(),
        name='exclude_path_regexes',
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        name='scan_timeout',
        doc='consider scan as failed if scan time exceeds timeout (in seconds)',
        default=3600,
        type=int,
    ),
)


FILTER_ATTRIBUTES = (
    AttributeSpec.required(
        name='type',
        doc='defines type to apply filter on, (component|source|resource)',
        type=str,
    ),
    AttributeSpec.required(
        name='match',
        doc='matches artifacts (list of regex|true|false)',
        type=typing.Union[dict, bool],
    ),
    AttributeSpec.required(
        name='action',
        doc='defines action to matched artifacts (include|exclude)',
        type=str,
    ),
)


class FilterCfg(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return FILTER_ATTRIBUTES

    def type(self):
        return self.raw['type']

    def match(self):
        return self.raw['match']

    def action(self):
        return self.raw['action']


class CheckmarxCfg(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return CHECKMARX_ATTRIBUTES

    def team_id(self):
        return self.raw['team_id']

    def severity_threshold(self) -> int:
        return self.raw.get('severity_threshold')

    def checkmarx_cfg_name(self):
        return self.raw.get('cfg_name')

    def include_path_regexes(self):
        return self.raw['include_path_regexes']

    def exclude_path_regexes(self):
        return self.raw['exclude_path_regexes']

    def scan_timeout(self) -> int:
        return self.raw.get('scan_timeout')


ATTRIBUTES = (
    AttributeSpec.optional(
        name='notify',
        default=Notify.EMAIL_RECIPIENTS,
        doc='whom to notify about found issues',
        type=Notify,
    ),
    AttributeSpec.optional(
        name='email_recipients',
        default=(),
        doc='optional list of email recipients to be notified about critical scan results',
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        name='filters',
        default={},
        doc='config to include and exclude sources, resources or whole components',
        type=FilterCfg,
    ),
    AttributeSpec.optional(
        name='checkmarx',
        type=CheckmarxCfg,
        default=(),
        doc='if present, perform checkmarx scanning',
    ),
    AttributeSpec.optional(
        name='issue_policies',
        default=IssuePolicies(),
        type=IssuePolicies,
        doc='defines issues policies (e.g. SLAs for maximum processing times',
    ),
    AttributeSpec.optional(
        name='overwrite_github_issues_tgt_repository_url',
        default=None,
        doc='if set, and notify is set to github_issues, overwrite target github repository',
    ),
    AttributeSpec.optional(
        name='github_issue_templates',
        default=None,
        doc='''\
        use to configure custom github-issue-templates (sub-attr: `body`)
        use python3's format-str syntax
        available variables:
        - summary # contains name, version, etc in a table
        - component_name
        - component_version
        - resource_name
        - resource_version
        - resource_type
        - greatest_cve
        - report_url
        - delivery_dashboard_url
        ''',
        type=list[GithubIssueTemplateCfg],
    ),
    AttributeSpec.optional(
        name='github_issue_labels_to_preserve',
        default=None,
        doc='optional list of regexes for labels that will never be removed upon ticket-update',
        type=list[str],
    ),

)


class SourceScanTrait(Trait):
    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def _children(self):
        if self.checkmarx():
            yield self.checkmarx()

    def notify(self):
        return Notify(self.raw['notify'])

    def email_recipients(self):
        return self.raw['email_recipients']

    def checkmarx(self):
        if checkmarx := self.raw.get('checkmarx'):
            return CheckmarxCfg(checkmarx)

    def filters(self):
        if filters := self.raw.get('filters'):
            return FilterCfg(filters)

    def filters_raw(self):
        if filters := self.raw.get('filters'):
            return filters

    def transformer(self):
        return SourceScanTraitTransformer(trait=self)

    def custom_init(self, raw_dict: dict):
        if self.checkmarx():
            return True
        else:
            # TODO should actually raise something, but breaks docu generation
            ci.util.warning('checkmarx should be defined.')

    def issue_policies(self) -> IssuePolicies:
        if isinstance((v := self.raw['issue_policies']), IssuePolicies):
            return v

        return dacite.from_dict(
            data_class=IssuePolicies,
            data=v,
        )

    def overwrite_github_issues_tgt_repository_url(self) -> typing.Optional[str]:
        return self.raw.get('overwrite_github_issues_tgt_repository_url')

    def github_issue_templates(self) -> list[GithubIssueTemplateCfg]:
        if not (raw := self.raw.get('github_issue_templates')):
            return None

        template_cfgs = [
            dacite.from_dict(
                data_class=GithubIssueTemplateCfg,
                data=cfg,
            ) for cfg in raw
        ]

        return template_cfgs

    def github_issue_template(self, type: str) -> typing.Optional[GithubIssueTemplateCfg]:
        if not (template_cfgs := self.github_issue_templates()):
            return None

        for cfg in template_cfgs:
            if cfg.type == type:
                return cfg

        return None

    def github_issue_labels_to_preserve(self) -> typing.Optional[list[str]]:
        return self.raw['github_issue_labels_to_preserve']


class SourceScanTraitTransformer(TraitTransformer):
    name = 'scan_sources'

    def __init__(self, trait, *args, **kwargs):
        self.trait = trait
        super().__init__(*args, **kwargs)

    def inject_steps(self):
        self.source_scan_step = PipelineStep(
            name='scan_sources',
            raw_dict={},
            is_synthetic=True,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3
        )
        self.source_scan_step.add_input(
            name=concourse.model.traits.component_descriptor.DIR_NAME,
            variable_name=concourse.model.traits.component_descriptor.ENV_VAR_NAME,
        )
        self.source_scan_step.set_timeout(duration_string='18h')
        yield self.source_scan_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # our step depends on dependency descriptor step
        component_descriptor_step = pipeline_args.step(
            concourse.model.traits.component_descriptor.DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME
        )
        self.source_scan_step._add_dependency(component_descriptor_step)

    @classmethod
    def dependencies(cls):
        return {'component_descriptor'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
    StepNotificationPolicy,
)
from concourse.model.base import (
  AttributeSpec,
  ModelValidationError,
  ScriptType,
  Trait,
  TraitTransformer,
)
from concourse.model.job import (
  JobVariant,
)

ATTRIBUTES = (
    AttributeSpec.optional(
        name='preprocess',
        default='finalize',
        doc='version processing operation to set effective version',
    ),
)


class DraftReleaseTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def transformer(self):
        return DraftReleaseTraitTransformer()

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def _preprocess(self):
        return self.raw['preprocess']

    def validate(self):
        super().validate()
        if self._preprocess() != 'finalize':
            raise ModelValidationError(
                "Only 'finalize' is supported as value for 'preprocess' in draft_release trait"
            )


class DraftReleaseTraitTransformer(TraitTransformer):
    name = 'draft_release'

    def inject_steps(self):
        # inject 'release' step
        self.draft_release_step = PipelineStep(
            name='create_draft_release_notes',
            raw_dict={},
            is_synthetic=True,
            notification_policy=StepNotificationPolicy.NO_NOTIFICATION,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
        )
        self.draft_release_step.set_timeout(duration_string='2h')
        yield self.draft_release_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        cd_trait = pipeline_args.trait('component_descriptor')
        cd_step = pipeline_args.step(cd_trait.step_name())
        self.draft_release_step._add_dependency(cd_step)

    @classmethod
    def dependencies(cls):
        return {'version', 'component_descriptor'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import functools

from ci.util import not_none

from concourse.model.base import ModelValidationError


@functools.cache
def _traits():
    from .component_descriptor import ComponentDescriptorTrait
    from .cronjob import CronTrait
    from .draft_release import DraftReleaseTrait
    from .image_scan import ImageScanTrait
    from .notifications import NotificationsTrait
    from .options import OptionsTrait
    from .publish import PublishTrait
    from .pullrequest import PullRequestTrait
    from .release import ReleaseTrait
    from .scheduling import SchedulingTrait
    from .slack import SlackTrait
    from .update_component_deps import UpdateComponentDependenciesTrait
    from .version import VersionTrait
    from .scan_sources import SourceScanTrait
    TRAITS = {
        'component_descriptor': ComponentDescriptorTrait,
        'cronjob': CronTrait,
        'draft_release': DraftReleaseTrait,
        'image_scan': ImageScanTrait,
        'notifications': NotificationsTrait,
        'options': OptionsTrait,
        'publish': PublishTrait,
        'pull-request': PullRequestTrait,
        'release': ReleaseTrait,
        'scheduling': SchedulingTrait,
        'slack': SlackTrait,
        'update_component_deps': UpdateComponentDependenciesTrait,
        'version': VersionTrait,
        'scan_sources': SourceScanTrait,
    }

    return TRAITS


class TraitsFactory:
    @staticmethod
    def create(
        name: str,
        variant_name: str,
        args_dict: dict,
        cfg_set,
    ):
        if name not in _traits():
            raise ModelValidationError('no such trait: ' + str(name))
        not_none(args_dict)

        ctor = _traits()[name]

        return ctor(
            name=name,
            variant_name=variant_name,
            raw_dict=args_dict,
            cfg_set=cfg_set,
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import logging

import git
import yaml

import ci.util
import cnudie.util
import cnudie.retrieve
import cnudie.migrate
import version

import ocm

logger = logging.getLogger('step.component_descriptor')


def dump_component_descriptor_v2(component_descriptor_v2: ocm.ComponentDescriptor):
    import ocm
    import dataclasses
    import yaml
    return yaml.dump(
        data=dataclasses.asdict(component_descriptor_v2),
        Dumper=ocm.EnumValueYamlDumper,
    )


def base_component_descriptor_v2(
    component_name_v2: str,
    component_labels: list[ocm.Label],
    effective_version: str,
    source_labels: tuple,
    ctx_repository_base_url: str,
    commit: str,
    repo_url: str,
):
    import datetime
    import cnudie.migrate
    import ocm
    import version as version_util
    parsed_version = version_util.parse_to_semver(effective_version)
    if parsed_version.finalize_version() == parsed_version:
        # "final" version --> there will be a tag, later (XXX hardcoded hack)
        src_ref = f'refs/tags/{effective_version}'
    elif parsed_version.prerelease.startswith('timestamp'):
        # prerelease is the build timestamp, not a "real" ref
        src_ref = None
    else:
        # let's hope the version contains something committish
        if parsed_version.build:
            src_ref = f'{parsed_version.prerelease}{parsed_version.build}'
        else:
            src_ref = f'{parsed_version.prerelease}'

    # logical names must not contain slashes or dots
    logical_name = component_name_v2.replace('/', '_').replace('.', '_')

    if cnudie.migrate.emit_compatible_to_cnudie_v2:
        provider = 'internal'
    else:
        provider = {
            'name': 'SAP SE',
        }

    component_labels = list(component_labels)
    component_labels.append(
        ocm.Label(
            name='cloud.gardener/ocm/creation-date',
            value=datetime.datetime.now(tz=datetime.timezone.utc).isoformat(),
        ),
    )

    base_descriptor_v2 = ocm.ComponentDescriptor(
      meta=ocm.Metadata(schemaVersion=ocm.SchemaVersion.V2),
      component=ocm.Component(
        name=component_name_v2,
        version=effective_version,
        repositoryContexts=[
          ocm.OciOcmRepository(
            baseUrl=ctx_repository_base_url,
            type=ocm.AccessType.OCI_REGISTRY,
          )
        ],
        provider=provider,
        sources=[
          ocm.Source(
            name=logical_name,
            type=ocm.ArtefactType.GIT,
            access=ocm.GithubAccess(
              type=ocm.AccessType.GITHUB,
              repoUrl=repo_url,
              ref=src_ref,
              commit=commit,
            ),
            version=effective_version,
            labels=source_labels,
          )
        ],
        componentReferences=[], # added later
        resources=[], # added later
        labels=component_labels,
        creationTime=datetime.datetime.now(datetime.timezone.utc).isoformat(),
      ),
    )

    return base_descriptor_v2


def component_diff_since_last_release(
    component_descriptor: ocm.ComponentDescriptor,
    component_descriptor_lookup,
    version_lookup,
):
    component: ocm.Component = ci.util.not_none(
        component_descriptor.component,
    )

    versions = version_lookup(
        component.identity()
    )

    greatest_release_version = version.greatest_version_before(
        reference_version=component.version,
        versions=versions,
        ignore_prerelease_versions=True,
    )

    if not greatest_release_version:
        logger.warning('could not determine last release version')
        return None

    greatest_release_version = str(greatest_release_version)
    logger.info('last released version: ' + str(greatest_release_version))

    greatest_released_cd = component_descriptor_lookup(ocm.ComponentIdentity(
        name=component.name,
        version=greatest_release_version,
    ))
    greatest_released_component = greatest_released_cd.component

    if not greatest_released_component:
        ci.util.fail(f'could not find {component=}')

    return cnudie.retrieve.component_diff(
        left_component=component,
        right_component=greatest_released_component,
        ignore_component_names=(component.name,),
        component_descriptor_lookup=component_descriptor_lookup,
    )


def write_component_diff(component_diff, out_path):
    # let us write only a subset for now, namely component names with changed versions
    diff_dict = {
        'component_names_with_version_changes': list(component_diff.names_version_changed),
    }

    with open(out_path, 'w') as f:
        yaml.dump(diff_dict, f)


def head_commit_hexsha(repo_path):
    git_repo = git.Repo(repo_path)
    if not git_repo.head.is_valid():
        commit_hash = None
    else:
        try:
            commit_hash = git_repo.head.commit.hexsha
        except:
            commit_hash = None

    return commit_hash



import dacite

import model.base


class GcpServiceAccount(model.base.NamedModelElement):
    def service_account_key(self):
        '''
        service-account-key (credentials) as retrieved from GCP's IAM & Admin console
        '''
        return self.raw['service_account_key']

    def service_account_credentials(self) -> 'google.oauth2.service_account.Credentials':
        import google.oauth2.service_account
        return google.oauth2.service_account.Credentials.from_service_account_info(
            self.service_account_key()
        )

    def project(self):
        return self.raw['project']

    def client_email(self) -> str:
        return self.service_account_key()['client_email']

    def private_key_id(self) -> str:
        return self.service_account_key()['private_key_id']

    def rotation_cfg(self) -> model.base.CfgElementReference:
        '''
        used to specify cfg-element to use for cross-rotation
        '''
        raw = self.raw.get('rotation_cfg')
        if raw:
            return dacite.from_dict(
                data_class=model.base.CfgElementReference,
                data=raw,
            )

        return None

    def _required_attributes(self):
        return ['service_account_key','project']



import tarfile

import dacite

import version
import unixutil.model as um


def _parse_os_release(contents: str):
    for line in contents.split('\n'):
        line = line.strip()
        if not line or line.startswith('#'):
            continue

        name, value = line.split('=', 1)

        yield (name, value.strip('"'))


def _parse_centos_release(contents: str):
    contents = contents.strip()
    lines = [line for line in contents.split('\n') if line]

    if len(lines) != 1:
        raise ValueError('expected a single line')

    line = lines[0]

    # expected format: "CentOS Linux release <version> (Core)"
    parts = line.split(' ')
    version = parts[3]

    yield ('VERSION_ID', version)


def _parse_debian_version(contents: str):
    contents = contents.strip()
    lines = [line for line in contents.split('\n') if line]

    if len(lines) != 1:
        raise ValueError('expected a single line')

    # file is expected to contain exactly the version
    yield ('VERSION_ID', lines[0])


def determine_osinfo(tarfh: tarfile.TarFile) -> um.OperatingSystemId | None:
    '''
    tries to determine the operating system identification, roughly as specified by
        https://www.freedesktop.org/software/systemd/man/os-release.html
    and otherwise following some conventions believed to be common.

    The argument (an opened tarfile) is being read from its initial position, possibly (but
    not necessarily) to the end. The underlying stream does not need to be seekable.
    It is the caller's responsibility to close the tarfile handle after this function returns.

    The tarfile is expected to contain a directory tree from a "well-known" unix-style operating
    system distribution. In particular, the following (GNU/) Linux distributions are well-supported:
    - alpine
    - debian
    - centos

    In case nothing was recognised within the given tarfile, the returned OperatingSystemId's
    attributes will all be `None`.
    '''
    known_fnames = (
        'debian_version',
        'centos-release',
        'os-release',
    )

    os_info = {}

    for info in tarfh:
        fname = info.name.split('/')[-1]

        if not fname in known_fnames:
            continue

        if info.issym():
            # we assume fnames are the same (this assumption might not always be correct)
            continue

        if not info.isfile():
            continue

        # found an "interesting" file
        contents = tarfh.extractfile(info).read().decode('utf-8')

        if fname == 'os-release':
            for k,v in _parse_os_release(contents):
                if k in os_info:
                    if k == 'VERSION_ID' and version.is_semver_parseable(v) and \
                        not version.is_semver_parseable(os_info[k]):
                        pass
                    else:
                        continue # os-release has lesser precedence
                os_info[k] = v
            if os_info.get('ID') == 'ubuntu' and (ver := os_info.get('VERSION')):
                # of _course_ ubuntu requires a special hack
                os_info['VERSION_ID'] = ver.split(' ', 1)[0]
        elif fname == 'centos-release':
            for k,v in _parse_centos_release(contents):
                os_info[k] = v
        elif fname == 'debian_version':
            for k,v in _parse_debian_version(contents):
                if k in os_info:
                    if not version.is_semver_parseable(v):
                        continue # e.g. ubuntu has "misleading" debian_version
                os_info[k] = v
        else:
            raise NotImplementedError(fname)

    if not os_info:
        return None

    return dacite.from_dict(
        data_class=um.OperatingSystemId,
        data=os_info,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    NamedModelElement,
)


class CtxRepositoryCfg(NamedModelElement):
    '''
    a ctx-repository cfg, identifying a component-descriptor (v2) repository, defining a
    repository context.
    '''
    def base_url(self):
        return self.raw.get('base_url')

    def description(self):
        return self.raw.get('description', '<no description available>')

    def _required_attributes(self):
        return 'base_url',

    def _optional_attributes(self):
        return [
            'description',
        ]



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import dataclasses
import enum
from os import path
import typing

import yaml
import git
import tempfile
from urllib.parse import urlparse, ParseResult
import re
import logging

import ctx
import ocm
import ci.util

logger = logging.getLogger(__name__)


class BOMEntryType(enum.Enum):
    Docker = 'docker'


@dataclasses.dataclass
class BOMEntry:
    url: str
    type: BOMEntryType
    # component name
    comp: str
    mode: str = 'PREFER_MULTIARCH'


@dataclasses.dataclass
class ProductVersion:
    # product_version_id
    pv: str
    distributed: bool
    release_strategy: str
    # service_pack
    sp: str
    # patch_level
    pl: str
    name: str
    bom: typing.Sequence[BOMEntry] = ()


@dataclasses.dataclass
class RBSCBom:
    bomvers: str
    name: str
    # product_id
    p: str
    # product_versions
    pvs: typing.Sequence[ProductVersion] = ()


def injectCredentialsIntoRepoUrl(repo_url):
    ctx_factory = ctx.cfg_factory()
    gitlab = ctx_factory.gitlab("gitlab_rbsc_dev")
    credentials = gitlab.credentials()
    credentials_str = ':'.join((credentials.username(), credentials.passwd()))
    parsed_url = urlparse(repo_url)
    new = ParseResult(
        scheme=parsed_url.scheme,
        netloc=f'{credentials_str}@{parsed_url.netloc}',
        path=parsed_url.path,
        params=parsed_url.params,
        query=parsed_url.query,
        fragment=parsed_url.fragment,
    )
    return new.geturl()


def buildAndApplyBOM(
        bom_repository_url: str,
        bom_branch: str,
        bom_entries: typing.Sequence[BOMEntry],
    ):
    if bom_branch not in ["master", "val", "dev"]:
        raise ValueError("--rbsc-git-branch has to to be master, dev or val!")

    bom_repository_url = injectCredentialsIntoRepoUrl(bom_repository_url)

    # for the RBSC, it expects all urls to have the scheme and port explicitly specified.
    # Defaulting schema to https and port to 443 if not otherwise set.
    def normalizeUrlsWithSchemaAndPort(bom_entry: BOMEntry):
        url = bom_entry.url

        # urlparse requires the a protocol as officially specified. Prepend https if not set.
        if not re.search(r'^[A-Za-z0-9+.\-]+://', bom_entry.url):
            url = 'https://{0}'.format(bom_entry.url)

        parsed_url = urlparse(url)
        new = ParseResult(
            scheme=parsed_url.scheme,
            netloc=f'{parsed_url.hostname}:{parsed_url.port if parsed_url.port else 443}',
            path=parsed_url.path,
            params=parsed_url.params,
            query=parsed_url.query,
            fragment=parsed_url.fragment,
        )
        new_url = new.geturl()
        bom_entry.url = new_url
        return bom_entry

    with tempfile.TemporaryDirectory() as local_path:
        repo, origin = _pullGitForBranch(bom_repository_url, bom_branch, local_path)
        bom = _parseBOM(local_path)
        if len(bom.pvs) != 1:
            raise NotImplementedError("More than one product version is not supported right now")

        normalized_boms = list(map(normalizeUrlsWithSchemaAndPort, bom_entries))

        # sort by name to have a reproducable bom with good diff highlighting
        normalized_sorted_boms = sorted(normalized_boms, key=lambda entry: entry.comp)

        # Deduplicate bom list based on url (since rbsc does not support two entries with
        # the same url)
        deduplicated_bom_list = []
        preexisting_urls = set()
        for bom_entry in normalized_sorted_boms:
            if bom_entry.url not in preexisting_urls:
                deduplicated_bom_list.append(bom_entry)
                preexisting_urls.add(bom_entry.url)
        bom.pvs[0].bom = deduplicated_bom_list

        _writeBOM(bom, local_path)
        _pushToGit(repo, origin)


def _parseBOM(local_path: str) -> RBSCBom:
    with open(path.join(local_path, "bom.yaml"), "r") as bom_file:
        raw_src_bom = yaml.load(bom_file, Loader=yaml.SafeLoader)
        ci.util._count_elements(raw_src_bom)
        src_bom = RBSCBom(
            bomvers=raw_src_bom.get("bomvers"),
            name=raw_src_bom.get("name"),
            p=raw_src_bom.get("p"),
            pvs=[ProductVersion(
                pv=pv.get("pv"),
                distributed=pv.get("distributed"),
                release_strategy=pv.get("release_strategy"),
                sp=pv.get("sp"),
                pl=pv.get("pl"),
                name=pv.get("name"),
                bom=[
                    BOMEntry(
                        comp=entry.get("comp"),
                        type=entry.get("type"),
                        url=entry.get("url"),
                    ) for entry in pv.get("bom")
                ],
            ) for pv in raw_src_bom.get("pvs")],
        )
        return src_bom


def _writeBOM(bom: RBSCBom, local_path: str):
    with open(path.join(local_path, "bom.yaml"), 'w') as bom_file:
        yaml.dump(
            data=dataclasses.asdict(bom),
            stream=bom_file,
            Dumper=ocm.EnumValueYamlDumper,
        )


def _pullGitForBranch(bom_repository_url: str, bom_branch: str, local_path: str):
    repo = git.Repo.init(local_path, mkdir=True)

    # setup remote
    origin = repo.create_remote('origin', bom_repository_url)
    origin.fetch()

    remote_ref = origin.refs[bom_branch]
    if remote_ref is None:
        raise RuntimeError(f"Remote Branch for {bom_branch} does not exist!")
    logging.info(
        f'Git pull repo with bom.yaml from {origin.name}...'
    )
    repo.create_head(bom_branch, remote_ref).set_tracking_branch(remote_ref).checkout()
    origin.pull()
    logging.info(
        f'Git pull repo with bom.yaml from {origin.name} completed.'
    )
    return repo, origin


def _pushToGit(repo: git.Repo, origin: git.Remote):
    logging.info(
        f'Git push repo with bom.yaml to {origin.name}...'
    )

    actor = git.Actor("CNUDIE-Transport-Tool", "")

    index = repo.index
    index.add([path.join(repo.working_tree_dir, 'bom.yaml')])
    index.commit("New BOM Version", author=actor, committer=actor)
    push_response = origin.push()

    if len(push_response) == 0:
        logging.error(
            f'Git push repo with bom.yaml to {origin.name} failed.'
        )

    failed_heads = [r for r in push_response if r.flags == git.remote.PushInfo.ERROR]
    if len(failed_heads) > 0:
        logging.error(
            f'Git push repo with bom.yaml to {origin.name} failed for heads {failed_heads}.'
        )
    else:
        logging.info(
                f'Git push repo with bom.yaml to {origin.name} completed.'
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import hashlib
import json
import logging
import tarfile
import tempfile
import typing
import zlib

import requests

import ccc.oci
import gziputil
import oci
import oci.client as oc
import oci.convert as oconv
import oci.platform
import oci.model as om
import tarutil

logger = logging.getLogger(__name__)


def image_exists(image_reference: str):
    oci_client = ccc.oci.oci_client()
    return bool(oci_client.head_manifest(image_reference=image_reference, absent_ok=True))


def filter_image(
    source_ref: typing.Union[str, om.OciImageReference],
    target_ref: typing.Union[str, om.OciImageReference],
    remove_files: typing.Sequence[str]=(),
    oci_client: oc.Client=None,
    mode: oci.ReplicationMode=oci.ReplicationMode.REGISTRY_DEFAULTS,
    platform_filter: typing.Callable[[om.OciPlatform], bool]=None,
    oci_manifest_annotations: dict[str, str]=None,
) -> typing.Tuple[requests.Response, str, bytes]: # response, tgt-ref, manifest_bytes
    if not oci_client:
        oci_client = ccc.oci.oci_client()

    source_ref = om.OciImageReference.to_image_ref(source_ref)
    target_ref = om.OciImageReference.to_image_ref(target_ref)

    # shortcut in case there are no filtering-rules
    if not remove_files:
        return oci.replicate_artifact(
            src_image_reference=source_ref,
            tgt_image_reference=target_ref,
            oci_client=oci_client,
            mode=mode,
            platform_filter=platform_filter,
            annotations=oci_manifest_annotations,
        )

    if mode is oci.ReplicationMode.REGISTRY_DEFAULTS:
        accept = None
    elif mode is oci.ReplicationMode.PREFER_MULTIARCH:
        accept = om.MimeTypes.prefer_multiarch
    elif mode is oci.ReplicationMode.NORMALISE_TO_MULTIARCH:
        accept = om.MimeTypes.prefer_multiarch
    else:
        raise NotImplementedError(mode)

    manifest = oci_client.manifest(
        image_reference=str(source_ref),
        accept=accept,
    )

    if isinstance(manifest, om.OciImageManifestList):
        # recurse into sub-images

        src_name = source_ref.ref_without_tag
        tgt_name = target_ref.ref_without_tag

        patched_manifests = []
        for sub_manifest in tuple(manifest.manifests):
            source_ref = f'{src_name}@{sub_manifest.digest}'

            if platform_filter:
                platform = oci.platform.from_single_image(
                    image_reference=source_ref,
                    oci_client=oci_client,
                    base_platform=sub_manifest.platform,
                )
                if not platform_filter(platform):
                    logger.info(f'skipping {platform=} for {source_ref=}')
                    manifest.manifests.remove(sub_manifest)
                    continue

            logger.info(f'filtering to {tgt_name=}')

            res, tgt_ref, manifest_bytes = filter_image(
                source_ref=source_ref,
                target_ref=tgt_name,
                remove_files=remove_files,
                oci_client=oci_client,
                oci_manifest_annotations=oci_manifest_annotations,
            )

            # patch (potentially) modified manifest-digest
            patched_manifest = dataclasses.replace(
                sub_manifest,
                digest=f'sha256:{hashlib.sha256(manifest_bytes).hexdigest()}',
                size=len(manifest_bytes),
            )
            patched_manifests.append(patched_manifest)

        manifest.manifests = patched_manifests
        manifest_dict = manifest.as_dict()
        manifest_raw = json.dumps(manifest_dict).encode('utf-8')
        res = oci_client.put_manifest(
            image_reference=str(target_ref),
            manifest=manifest_raw,
        )

        return res, str(target_ref), manifest_raw

    # normalise single-image to multi-arch (w/ one entry)
    if mode is oci.ReplicationMode.NORMALISE_TO_MULTIARCH:
        if not source_ref.has_digest_tag:
            source_ref = om.OciImageReference.to_image_ref(
                oci_client.to_digest_hash(
                    image_reference=source_ref,
                )
            )

        platform = oci.platform.from_single_image(
            image_reference=source_ref,
            oci_client=oci_client,
        )

        res, ref, manifest_bytes = filter_image(
            source_ref=source_ref,
            target_ref=target_ref.ref_without_tag,
            remove_files=remove_files,
            oci_client=oci_client,
            oci_manifest_annotations=oci_manifest_annotations,
        )

        manifest_list = om.OciImageManifestList(
            manifests=[
                om.OciImageManifestListEntry(
                    digest=f'sha256:{hashlib.sha256(manifest_bytes).hexdigest()}',
                    mediaType=manifest.mediaType,
                    size=len(manifest_bytes),
                    platform=platform,
                )
            ],
        )

        manifest_list_bytes = json.dumps(
            manifest_list.as_dict(),
        ).encode('utf-8')

        res = oci_client.put_manifest(
            image_reference=target_ref,
            manifest=manifest_list_bytes,
        )

        return res, target_ref, manifest_list_bytes

    cp_cfg_blob = True
    if isinstance(manifest, om.OciImageManifestV1):
        logger.info(f'converting v1-manifest -> v2 {source_ref=} {target_ref=}')
        manifest, cfg_blob = oconv.v1_manifest_to_v2(
            manifest=manifest,
            oci_client=oci_client,
            tgt_image_ref=str(target_ref),
        )
        cp_cfg_blob = False # we synthesise new cfg - thus we cannot cp from src
    elif not isinstance(manifest, om.OciImageManifest):
        raise NotImplementedError(manifest)

    # allow / ignore leading '/'
    remove_files = [p.lstrip('/') for p in remove_files]

    def tarmember_filter(tar_info: tarfile.TarInfo):
        stripped_name = tar_info.name.lstrip('./')
        if stripped_name in remove_files:
            logger.debug(f'rm: {tar_info.name=}')
            return False # rm member
        return True # keep member

    # prepare copy of layers to avoid modification while iterating
    layers_copy = manifest.layers.copy()

    non_gzipped_layer_digests = {} # {gzipped-digest: sha256:non-gzipped-digest}

    have_non_tar_layer = False
    patch_cfg_blob = True
    for layer in manifest.layers:
        layer_hash = hashlib.sha256()
        cfg_hash = hashlib.sha256() # we need to write "non-gzipped" hash to cfg-blob
        leng = 0
        src_leng = 0 # required for calculating leng for gzip-footer
        crc = 0 # requried for calculcating crc32-checksum for gzip-footer

        if not 'tar' in layer.mediaType:
            have_non_tar_layer = True
            cp_cfg_blob = True
            patch_cfg_blob = False

            # special-case: do not filter "layer", if it is not a tar (e.g. the case for
            # "in-toto" (application/vnd.in-toto+json)
            if oci_client.head_blob(
                image_reference=target_ref,
                digest=layer.digest,
                absent_ok=True,
            ):
                continue # skip blob replication if already present in tgt

            blob = oci_client.blob(
                image_reference=str(source_ref),
                digest=layer.digest,
                stream=True,
            )
            oci_client.put_blob(
                image_reference=target_ref,
                digest=layer.digest,
                octets_count=layer.size,
                data=blob,
            )
            continue

        if have_non_tar_layer:
            raise RuntimeError(
                'don\'t know how to process mixed image (tar + non-tar layers)'
            )

        # unfortunately, GCR (our most important oci-registry) does not support chunked uploads,
        # so we have to resort to writing the streaming result into a local tempfile to be able
        # to calculate digest-hash prior to upload to tgt; XXX: we might use streaming
        # when interacting w/ oci-registries that support chunked-uploads
        with tempfile.TemporaryFile() as f:
            src_tar_stream = oci_client.blob(
                image_reference=str(source_ref),
                digest=layer.digest,
                stream=True,
            ).iter_content(chunk_size=tarfile.BLOCKSIZE * 64)
            src_tar_fobj = tarutil.FilelikeProxy(generator=src_tar_stream)
            filtered_stream = tarutil.filtered_tarfile_generator(
                src_tf=tarfile.open(fileobj=src_tar_fobj, mode='r|*'),
                filter_func=tarmember_filter,
                chunk_size=tarfile.BLOCKSIZE * 64,
            )

            f.write((gzip_header := gziputil.gzip_header(fname=b'layer.tar')))
            layer_hash.update(gzip_header)
            leng += len(gzip_header)

            compressor = gziputil.zlib_compressobj()

            for chunk in filtered_stream:
                cfg_hash.update(chunk) # need to hash before compressing for cfg-blob
                crc = zlib.crc32(chunk, crc)
                src_leng += len(chunk)

                chunk = compressor.compress(chunk)
                layer_hash.update(chunk)
                leng += len(chunk)
                f.write(chunk)

            f.write((remainder := compressor.flush()))
            layer_hash.update(remainder)
            leng += len(remainder)

            gzip_footer = gziputil.gzip_footer(
                crc32=crc,
                uncompressed_size=src_leng,
            )
            f.write(gzip_footer)
            layer_hash.update(gzip_footer)
            leng += len(gzip_footer)

            f.seek(0)

            oci_client.put_blob(
                image_reference=target_ref,
                digest=(layer_digest := 'sha256:' + layer_hash.hexdigest()),
                octets_count=leng,
                data=f,
            )

            non_gzipped_layer_digests[layer_digest] = 'sha256:' + cfg_hash.hexdigest()

            # update copy of layers-list with new layer
            new_layer = dataclasses.replace(layer, digest=layer_digest, size=leng)
            layers_copy[layers_copy.index(layer)] = new_layer

    # switch layers in manifest to announce changes w/ manifest-upload
    manifest.layers = layers_copy

    # need to patch cfg-object, in case layer-digests changed
    if cp_cfg_blob:
        cfg_blob = oci_client.blob(
            image_reference=str(source_ref),
            digest=manifest.config.digest,
            stream=False,
        ).content # cfg-blobs are small - no point in streaming
    else:
        cfg_blob = json.loads(cfg_blob)

    if patch_cfg_blob:
        if isinstance(cfg_blob, bytes):
            cfg_blob = json.loads(cfg_blob)

        if not 'rootfs' in cfg_blob:
            raise ValueError('expected attr `rootfs` not present on cfg-blob')
        cfg_blob['rootfs'] = {
            'diff_ids': [
                non_gzipped_layer_digests[layer.digest] for layer in manifest.layers
            ],
            'type': 'layers',
        }

    if isinstance(cfg_blob, dict):
        cfg_blob = json.dumps(cfg_blob).encode('utf-8')

    cfg_digest = f'sha256:{hashlib.sha256(cfg_blob).hexdigest()}'
    cfg_leng = len(cfg_blob)
    oci_client.put_blob(
        image_reference=str(target_ref),
        digest=cfg_digest,
        octets_count=cfg_leng,
        data=cfg_blob,
    )

    manifest.config = dataclasses.replace(manifest.config, digest=cfg_digest, size=cfg_leng)

    manifest_raw = json.dumps(manifest.as_dict()).encode('utf-8')

    if oci_manifest_annotations:
        manifest_dict = json.loads(manifest_raw)
        if not 'annotations' in manifest_dict:
            manifest_dict['annotations'] = {}

        manifest_dict['annotations'] |= oci_manifest_annotations

        manifest_raw = json.dumps(manifest_dict).encode('utf-8')

    if target_ref.has_tag:
        target_ref = str(target_ref)
    else:
        # if tgt does not bear a tag, calculate hash digest as tgt
        manifest_digest = hashlib.sha256(manifest_raw).hexdigest()
        target_ref = f'{target_ref.ref_without_tag}@sha256:{manifest_digest}'

    res = oci_client.put_manifest(
        image_reference=target_ref,
        manifest=manifest_raw
    )
    res.raise_for_status()

    return res, target_ref, manifest_raw



import git
import os
import pytest

import gitutil


@pytest.fixture
def git_repo(tmpdir):
    repo = git.Repo.init(tmpdir)

    repo.index.commit('first commit')

    return repo


@pytest.fixture
def git_helper(git_repo):
    return gitutil.GitHelper(
        repo=git_repo,
        github_cfg=None,
        github_repo_path=None,
    )


def test_index_to_commit(git_helper):
    repo = git_helper.repo
    repo_dir = repo.working_tree_dir

    original_head_commit = repo.head.commit

    commit = git_helper.index_to_commit('empty commit')

    assert commit.message == 'empty commit'

    # index_to_commit should not change index
    assert repo.head.commit == original_head_commit

    new_file = os.path.join(repo_dir, 'new_file.txt')
    with open(new_file, 'w') as f:
        f.write('new_file')

    commit = git_helper.index_to_commit('add new_file')

    # file should be left in working_tree (but not added to index)
    assert os.path.isfile(new_file)
    assert git_helper.is_dirty
    os.unlink(new_file)
    assert not git_helper.is_dirty

    repo.head.reset(commit, working_tree=True)
    # check "new_file.txt" was actually added to returned commit
    assert os.path.isfile(new_file)
    with open(new_file) as f:
        assert f.read() == 'new_file'

    # add "another" file to initial commit
    another_file = os.path.join(repo_dir, 'another_file')
    with open(another_file, 'w') as f:
        f.write('another_file')

    commit_with_another_file = git_helper.index_to_commit(
        message='add another file',
        parent_commits=(original_head_commit,)
    )

    assert commit_with_another_file.parents == [original_head_commit]

    # try the same, but pass commit-digest
    commit_with_another_file = git_helper.index_to_commit(
        message='add another file',
        parent_commits=(original_head_commit.hexsha,)
    )

    assert commit_with_another_file.parents == [original_head_commit]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import os

import concourse.model.base
import concourse.model.step
import concourse.model.job
import concourse.model.resources

# helper functions to create dummy pipeline definition objects


def pipeline_step(name):
    return concourse.model.step.PipelineStep(
        name=name,
        is_synthetic=True,
        script_type=concourse.model.base.ScriptType.PYTHON3,
        raw_dict={}
    )


def repository():
    return concourse.model.resources.RepositoryConfig(
        logical_name='main',
        raw_dict={
            'branch': 'master_branch',
            'path': 'org/repo_name',
        }
    )


def resource_registry():
    registry = concourse.model.resources.ResourceRegistry()
    # add dummy "main repository"
    return registry


def job(main_repo):
    job_variant = concourse.model.job.JobVariant(
        name='dummy_job',
        resource_registry=resource_registry(),
        raw_dict={},
    )

    job_variant._repos_dict = {main_repo.logical_name(): main_repo}
    job_variant._main_repository_name = main_repo.logical_name()
    job_variant._steps_dict = {}

    return job_variant


def populate_meta_dir(directory:str):
    for n in (
        'build-id',
        'build-name',
        'build-job-name',
        'build-team-name',
        'build-pipeline-name',
        'atc-external-url',
    ):
        with open(os.path.join(directory, n), 'w') as f:
            f.write(n)



import copy
import logging
import typing
import requests
import tempfile
import os
from contextlib import contextmanager
from dataclasses import dataclass
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import pkcs7

import cfg_mgmt
import cfg_mgmt.model as cmm
import ci.log
import ci.util
import model
import model.container_registry
from model.btp_application_certificate import BtpApplicationCertificate


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


class CertServiceClient:
    def __init__(
        self,
        credentials: dict,
    ):
        self.credentials = credentials
        self._setup_oauth_token()

    def _setup_oauth_token(self):
        uaa = self.credentials['uaa']
        data = {
            'grant_type': 'client_credentials',
            'token_format': 'bearer',
            'client_id': uaa['clientid'],
            'client_secret': uaa['clientsecret'],
        }
        headers = {
            'Accept': 'application/json',
        }
        resp = requests.post(
            f'{uaa["url"]}/oauth/token',
            data=data,
            headers=headers,
            timeout=(4, 31),
        )
        resp.raise_for_status()
        self.access_token = resp.json()['access_token']

    def create_client_certificate_chain(self, csr_pem: str, validity_in_days: int) -> dict:
        headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.access_token}',
        }
        data = {
            'csr':{
              'value': csr_pem,
            },
            'policy': self.credentials['certificateservice']['policy'],
            'validity': {'type': 'DAYS', 'value': validity_in_days},
        }
        url = self.credentials['certificateservice']['apiurl'] + '/v3/synchronous/certificate'
        resp = requests.post(
            url,
            json=data,
            headers=headers,
            timeout=(4, 31),
        )
        resp.raise_for_status()
        logger.info('Created certificate')
        return resp.json()['certificateChain']


def _write_temp_file(temp_dir: str, fname: str, content: str) -> str:
    fpath = os.path.join(temp_dir, fname)
    with open(fpath, 'w', encoding='utf-8') as f:
        f.write(content)
    return fpath


@dataclass
class CertInfo:
    id: str
    dn: str
    cn: str
    serial_no: int


class GBaasAppClient:
    def __init__(self, auth: BtpApplicationCertificate):
        endpoint = auth.application_endpoint()
        self.url = f'{endpoint}/service/sps/{auth.application_id()}/apiCertificate'
        self.clienturl = f'{endpoint}/service/sps/{auth.client_id()}/apiCertificate'
        self.certificate_pem = auth.certificate_pem()
        self.private_key_pem = auth.private_key_pem()

    @contextmanager
    def _session_with_cert(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            session = requests.Session()
            crt_fname = _write_temp_file(temp_dir, 'cert.pem', self.certificate_pem)
            key_fname = _write_temp_file(temp_dir, 'key.pem', self.private_key_pem)
            session.cert = (crt_fname, key_fname)
            yield session

    def put_certificate(self, cert_pem: str, desc: str, scopes: list[str]) -> str:
        cert_pem = cert_pem.strip()
        cert_pem = cert_pem.removeprefix("-----BEGIN CERTIFICATE-----\n")
        cert_pem = cert_pem.removesuffix("-----END CERTIFICATE-----")
        data = {
            'description': desc,
            'scopes': scopes,
            'base64': cert_pem,
        }
        with self._session_with_cert() as session:
            resp = session.put(
                self.url,
                json=data,
                timeout=(4, 31),
            )
        resp.raise_for_status()
        id = resp.json()['certificateId']
        logger.info(f'Added certificate {id}')
        return id

    def delete_certificate(self, common_name: str, cert_id: str):
        data = {
            'certificateId': cert_id,
        }
        with self._session_with_cert() as session:
            resp = session.delete(
                self.url,
                json=data,
                timeout=(4, 31),
            )
        resp.raise_for_status()
        logger.info(f'Deleted certificate {common_name} ({cert_id})')

    @staticmethod
    def _find_cn(dn: str) -> str:
        for part in dn.split(','):
            key, value = part.split('=')
            if key == 'CN':
                return value
        return None

    def list_certificates_by_base(
        self,
        common_name_base: str
    ) -> typing.Generator[CertInfo, None, None]:
        with self._session_with_cert() as session:
            resp = session.get(
                self.clienturl,
                timeout=(4, 31),
            )
        resp.raise_for_status()
        for item in resp.json():
            id = item['dnId']
            dn = item['dn']
            cn = GBaasAppClient._find_cn(dn)
            if cn:
                try:
                    serial_no, base = BtpApplicationCertificate.parse_serial_no_from_common_name(cn)
                    if common_name_base == base:
                        yield CertInfo(id=id, dn=dn, cn=cn, serial_no=serial_no)
                except ValueError:
                    pass


_str_to_names = {
    'C': NameOID.COUNTRY_NAME,
    'ST': NameOID.STATE_OR_PROVINCE_NAME,
    'O': NameOID.ORGANIZATION_NAME,
    'OU': NameOID.ORGANIZATIONAL_UNIT_NAME,
    'L': NameOID.LOCALITY_NAME,
    'CN': NameOID.COMMON_NAME,
    'EMAIL': NameOID.EMAIL_ADDRESS,
}


def _create_csr(subject: str) -> tuple[str, str]:
    logger.info(f'Creating CSR for {subject}')
    key = rsa.generate_private_key(public_exponent=65537, key_size=3072)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode('utf-8')
    attributes = []
    for part in subject.split(', '):
        name, value = part.split('=')
        oid = _str_to_names[name]
        attributes.append(x509.NameAttribute(oid, value))
    csr = x509.CertificateSigningRequestBuilder().subject_name(
        x509.Name(attributes)
    ).sign(key, hashes.SHA256())
    csr_pem = csr.public_bytes(serialization.Encoding.PEM).decode('utf-8')
    return csr_pem, key_pem


def _extract_client_certificate(cert_response: dict) -> str:
    pkcs7_pem = cert_response['value']
    certs = pkcs7.load_pem_pkcs7_certificates(pkcs7_pem.encode('utf-8'))
    if not certs:
        raise ValueError('no certificates found in response')
    all_certs = ""
    for c in certs:
        all_certs += c.public_bytes(serialization.Encoding.PEM).decode('utf-8')
    return all_certs


def rotate_cfg_element(
    cfg_element: BtpApplicationCertificate,
    cfg_factory: model.ConfigFactory,
) -> typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:
    gbaas_auth = cfg_factory.btp_application_certificate(cfg_element.auth_application_certificate())
    gbaas_client = GBaasAppClient(gbaas_auth)

    # calc next serial no
    cn = cfg_element.common_name()
    serial_no, base = BtpApplicationCertificate.parse_serial_no_from_common_name(cn)
    next_sn = serial_no + 1
    for info in gbaas_client.list_certificates_by_base(base):
        if info.serial_no >= next_sn:
            next_sn = info.serial_no + 1
    next_cn = f'{next_sn}.{base}'

    # create certificate
    csr_pem, key_pem = _create_csr(cfg_element.subject(next_cn))
    sb_auth = cfg_factory.btp_service_binding(cfg_element.cert_service_binding())
    cs_client = CertServiceClient(sb_auth.credentials())
    response = cs_client.create_client_certificate_chain(csr_pem, cfg_element.validity_in_days())
    cert_pem = _extract_client_certificate(response)

    # add certificate to GBaas application
    id = gbaas_client.put_certificate(
        cert_pem=cert_pem,
        desc=f'CN={next_cn}',
        scopes=cfg_element.scopes(),
    )

    secret_id = {'common_name': cn}
    raw_cfg = copy.deepcopy(cfg_element.raw)
    raw_cfg['certificate_pem'] = cert_pem
    raw_cfg['private_key_pem'] = key_pem
    raw_cfg['common_name'] = next_cn
    updated_elem = BtpApplicationCertificate(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )

    def revert():
        gbaas_client.delete_certificate(next_cn, id)

    return revert, secret_id, updated_elem


def delete_config_secret(
    cfg_element: BtpApplicationCertificate,
    cfg_queue_entry: cmm.CfgQueueEntry,
    cfg_factory: model.ConfigFactory,
) -> BtpApplicationCertificate | None:
    logger.info('Deleting old certificates')
    gbaas_auth = cfg_factory.btp_application_certificate(cfg_element.auth_application_certificate())
    gbaas_client = GBaasAppClient(gbaas_auth)
    cn = cfg_queue_entry.secretId['common_name']
    serial_no, base = BtpApplicationCertificate.parse_serial_no_from_common_name(cn)
    for info in gbaas_client.list_certificates_by_base(base):
        if info.serial_no < serial_no:
            gbaas_client.delete_certificate(info.cn, info.id)

    return None



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
import logging
import typing

from github3 import GitHub, GitHubEnterprise
from github3.exceptions import NotFoundError
import github3.orgs
import github3.repos.repo
import github3.search.user
import github3.users

from ci.util import existing_dir, existing_file, not_none
import ci.log

logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


class Username(str):
    pass


class Team(str):
    @property
    def org_name(self) -> str:
        return self.split('/')[0]

    @property
    def name(self) -> str:
        return self.split('/')[1]


class EmailAddress(str):
    pass


def parse_codeowner_entry(
    entry: str,
) -> Username | EmailAddress | Team:
    '''
    Parse codeowner entry to Username, Email or Team.
    Invalid entries return `None`.
    '''

    if '@' not in entry:
        logger.warning(f'invalid codeowners-entry: {entry}')
        return

    if not entry.startswith('@'):
        return EmailAddress(entry) # plain email address

    entry = entry.removeprefix('@')

    if '/' not in entry:
        return Username(entry.lower())

    else:
        return Team(entry)


def enumerate_codeowners_from_remote_repo(
    repo: github3.repos.repo.Repository,
    paths: typing.Iterable[str] = ('CODEOWNERS', '.github/CODEOWNERS', 'docs/CODEOWNERS'),
) -> typing.Generator[Username | Team | EmailAddress, None, None]:
    for path in paths:
        try:
            yield from (
                parse_codeowner_entry(entry)
                for entry in filter_codeowners_entries(
                    repo.file_contents(path=path).decoded.decode('utf-8').split('\n'),
                )
            )
        except NotFoundError:
            pass # ignore absent files


def enumerate_codeowners_from_file(
    file_path: str,
) -> typing.Generator[Username | Team | EmailAddress, None, None]:
    file_path = existing_file(file_path)
    with open(file_path) as f:
        yield from (
            parse_codeowner_entry(entry)
            for entry in filter_codeowners_entries(f.readlines())
        )


def enumerate_codeowners_from_local_repo(
    repo_dir: str,
    paths: typing.Iterable[str] = ('CODEOWNERS', '.github/CODEOWNERS', 'docs/CODEOWNERS'),
) -> typing.Generator[Username | Team | EmailAddress, None, None]:
    repo_dir = existing_dir(Path(repo_dir))
    if not repo_dir.joinpath('.git').is_dir():
        raise ValueError(f'not a git root directory: {repo_dir}')

    for path in paths:
        codeowners_file = repo_dir.joinpath(path)
        if codeowners_file.is_file():
            with open(codeowners_file) as f:
                yield from (
                    parse_codeowner_entry(entry)
                    for entry in filter_codeowners_entries(f.readlines())
                )


def filter_codeowners_entries(
    lines: typing.Iterable[str],
) -> typing.Generator[str, None, None]:
    '''
    returns a generator yielding parsed entries from */CODEOWNERS
    each entry may be one of
        - a github user name (with a leading @ character)
        - a github team name (leading @ character and exactly one / character (org/name))
        - an email address
    '''
    for line in lines:
        line = line.strip()
        if line.startswith('#'):
            continue
        # Yield tokens, ignoring the first (it is the path filter)
        yield from line.split()[1:]


def determine_email_address(
    github_user_name: str | Username,
    github_api: GitHub,
) -> EmailAddress | None:
    '''
    Return email address exposed for given user.
    `None` returned if either user not found, or no email address is exposed.
    '''
    not_none(github_user_name)
    try:
        user = github_api.user(github_user_name)
    except NotFoundError:
        logger.warning(f'failed to lookup {github_user_name=} {github_api._github_url=}')
        return None

    if not user.email:
        return None

    return EmailAddress(user.email)


def usernames_from_email_address(
    email_address: EmailAddress,
    gh_api: GitHub | GitHubEnterprise,
) -> typing.Generator[Username, None, None]:
    no_user = True
    for res in gh_api.search_users(query=f'{email_address} in:email'):
        no_user = False
        yield Username(res.user.login)

    if no_user:
        logger.warning(f'no user found for {email_address=}')


def resolve_team_members(
    team: Team,
    github_api: GitHub,
    absent_ok: bool=True,
) -> typing.Generator[Username, None, None]:
    '''
    Return generator yielding usernames resolved recursively from given team.
    If no team found for given team, no users are returned.
    '''
    try:
        organisation = github_api.organization(team.org_name)
        team = organisation.team_by_name(team.name)
        team: github3.orgs.Team
    except NotFoundError:
        logger.warning('failed to lookup team {t}'.format(t=team.name))
        if absent_ok:
            return

        raise

    yield from (
        Username(member.login)
        for member in team.members()
    )


def resolve_email_addresses(
    codeowners_entries: typing.Iterable[Username | EmailAddress | Team],
    github_api: GitHub,
) -> typing.Generator[EmailAddress, None, None]:
    '''
    Returns a generator yielding the resolved email addresses for the given iterable of
    github codeowners entries.
    Teams are resolved to Users recursively.
    Users are resolved to exposed email addresses.
    If no email address is exposed the User is skipped.
    '''
    unique_email_addresses = set()

    for codeowner_entry in codeowners_entries:
        if isinstance(codeowner_entry, EmailAddress):
            unique_email_addresses.add(codeowner_entry)
            continue

        if isinstance(codeowner_entry, Username):
            if (email_address := determine_email_address(
                github_user_name=codeowner_entry,
                github_api=github_api,
            )):
                unique_email_addresses.add(email_address)
                continue

        if isinstance(codeowner_entry, Team):
            for email_address in resolve_email_addresses(
                codeowners_entries=resolve_team_members(
                    team=codeowner_entry,
                    github_api=github_api,
                ),
                github_api=github_api,
            ):
                unique_email_addresses.add(email_address)
            continue

    yield from unique_email_addresses


def resolve_usernames(
    codeowners_entries: typing.Iterable[Username | EmailAddress | Team],
    github_api: GitHub,
) -> typing.Generator[Username, None, None]:
    '''
    Returns a generator yielding the resolved usernames for the given iterable of
    github codeowners entries.
    Teams are resolved to Users recursively.
    Emails are resolved to users.
    If no username is found for given email address, its skipped.
    '''
    unique_usernames = set()

    for codeowner_entry in codeowners_entries:
        if isinstance(codeowner_entry, Username):
            unique_usernames.add(codeowner_entry)
            continue

        if isinstance(codeowner_entry, EmailAddress):
            for username in usernames_from_email_address(
                email_address=codeowner_entry,
                gh_api=github_api,
            ):
                unique_usernames.add(username)

            continue

        if isinstance(codeowner_entry, Team):
            for username in resolve_usernames(
                codeowners_entries=resolve_team_members(
                    team=codeowner_entry,
                    github_api=github_api,
                ),
                github_api=github_api,
            ):
                unique_usernames.add(username)

            continue

    yield from unique_usernames



import mako
import os
import pytest

from unittest.mock import MagicMock

from concourse.model.resources import RepositoryConfig
from model.github import GithubConfig
from resources import mako_resource_dir


class TestGithubMakoResource:
    test_file = os.path.join(mako_resource_dir(), 'github.mako')

    @pytest.fixture()
    def cfg_set(self):
        def _cfg_set(
            gh_config_name="foo",
            ssh_url='ssh://test.foo',
            http_url='https://test.foo',
            api_url='made-up-api-url',
            disable_tls_validation=False,
            available_protocols=['https', 'ssh'],
            tu_username='made-up-name',
            tu_password='made-up-password',
            tu_private_key='first key line \n second key line \n third key line',
            tu_auth_token='made-up-token',
            tu_email_address='username@host.domain',
        ):
            cfg_set_mock = MagicMock()
            gh_cfg = GithubConfig(
                name=gh_config_name,
                raw_dict={
                    'purpose_labels': ['ci'],
                    'sshUrl': ssh_url,
                    'httpUrl': http_url,
                    'apiUrl': api_url,
                    'disable_tls_validation': disable_tls_validation,
                    'available_protocols': available_protocols,
                    'technical_users': [{
                            'username': tu_username,
                            'password': tu_password,
                            'privateKey': tu_private_key,
                            'authToken': tu_auth_token,
                            'emailAddress': tu_email_address,
                    }],
                },
                type_name='github',
            )
            cfg_set_mock.github.return_value = gh_cfg
            cfg_set_mock._cfg_elements = lambda cfg_type_name: [gh_cfg]
            return cfg_set_mock
        return _cfg_set

    @pytest.fixture()
    def repo_cfg(self):
        def _repo_cfg(
            path='organisation/reponame',
            branch='master',
            include_paths=['path/to/include'],
            exclude_paths=['path/to/exclude'],
            hostname='test.foo'
        ):
            return RepositoryConfig(raw_dict={
                'path': path,
                'branch': branch,
                'hostname': hostname,
                'trigger_paths': {
                    'include': include_paths,
                    'exclude': exclude_paths,
                }
            })
        return _repo_cfg

    @pytest.mark.parametrize('require_label', [None, 'some-label'])
    @pytest.mark.parametrize('include_paths', [[], ['path/to/include']])
    @pytest.mark.parametrize('exclude_paths', [[], ['path/to/exclude']])
    @pytest.mark.parametrize('available_protocols', [['ssh','https'],['https', 'ssh']])
    def test_pr_resource_contains_required_attributes(
        self,
        repo_cfg,
        cfg_set,
        require_label,
        include_paths,
        exclude_paths,
        available_protocols,
    ):
        examinee = mako.template.Template(filename=self.test_file)
        test_repo_cfg = repo_cfg(include_paths=include_paths, exclude_paths=exclude_paths)
        test_cfg_set = cfg_set(available_protocols=available_protocols)
        render_result = examinee.get_def('github_pr').render(
            test_repo_cfg,
            test_cfg_set,
            require_label,
        )

        required_attributes = [
            'base', 'uri', 'api_endpoint', 'skip_ssl_verification', 'access_token', 'no_ssl_verify',
            'private_key', 'username', 'password',
        ]
        if require_label:
            required_attributes.append('label')
        if include_paths:
            required_attributes.append('paths')
        if exclude_paths:
            required_attributes.append('ignore_paths')

        result_lines = render_result.splitlines()
        missing_attributes = []
        for attr in required_attributes:
            if not any([attr in line for line in result_lines]):
                missing_attributes.append(attr)
        if missing_attributes:
            pytest.fail(
                f'Required attributes are missing in render result: {",".join(missing_attributes)}'
            )

    @pytest.mark.parametrize('include_paths', [[], ['path/to/include']])
    @pytest.mark.parametrize('exclude_paths', [[], ['path/to/exclude']])
    @pytest.mark.parametrize('available_protocols', [['ssh','https'],['https', 'ssh']])
    def test_git_resource_contains_required_attributes(
        self,
        repo_cfg,
        cfg_set,
        include_paths,
        exclude_paths,
        available_protocols,
    ):
        examinee = mako.template.Template(filename=self.test_file)
        test_repo_cfg = repo_cfg(include_paths=include_paths, exclude_paths=exclude_paths)
        test_cfg_set = cfg_set(available_protocols=available_protocols)
        render_result = examinee.get_def('github_repo').render(
            test_repo_cfg,
            test_cfg_set,
        )

        required_attributes = [
            'branch', 'uri', 'disable_ci_skip', 'skip_ssl_verification', 'no_ssl_verify',
            'private_key', 'username', 'password',
        ]
        if include_paths:
            required_attributes.append('paths')
        if exclude_paths:
            required_attributes.append('ignore_paths')

        result_lines = render_result.splitlines()
        missing_attributes = []
        for attr in required_attributes:
            if not any([attr in line for line in result_lines]):
                missing_attributes.append(attr)
        if missing_attributes:
            pytest.fail(
                f'Required attributes are missing in render result: {",".join(missing_attributes)}'
            )



from copy import copy
import logging
import logging.config
import sys
import tempfile


_debug_handler = None


def setup_debug_handler(
    custom_format_string: str = '',
    print_thread_id: bool = False,
):
    global _debug_handler
    if not _debug_handler:
        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        handler = logging.FileHandler(filename=tmpfile.name)
        handler.setLevel(logging.DEBUG)
        _debug_handler = handler

    if custom_format_string:
        _debug_handler.setFormatter(CCFormatter(fmt=custom_format_string))
    else:
        _debug_handler.setFormatter(CCFormatter(
            fmt=default_fmt_string(print_thread_id=print_thread_id))
        )

    return _debug_handler


class CCFormatter(logging.Formatter):
    level_colors = {
        logging.DEBUG: lambda level_name:
        f'{Bcolors.BOLD}{Bcolors.BLUE}{level_name}{Bcolors.RESET_ALL}',
        logging.INFO: lambda level_name:
        f'{Bcolors.BOLD}{Bcolors.GREEN}{level_name}{Bcolors.RESET_ALL}',
        logging.WARNING: lambda level_name:
        f'{Bcolors.BOLD}{Bcolors.YELLOW}{level_name}{Bcolors.RESET_ALL}',
        logging.ERROR: lambda level_name:
        f'{Bcolors.BOLD}{Bcolors.RED}{level_name}{Bcolors.RESET_ALL}',
    }

    def color_level_name(self, level_name, level_number):
        def default(level_name):
            return str(level_name)

        func = self.level_colors.get(level_number, default)
        return func(level_name)

    def formatMessage(self, record):
        record_copy = copy(record)
        levelname = record_copy.levelname
        if sys.stdout.isatty():
            levelname = self.color_level_name(levelname, record_copy.levelno)
            if "color_message" in record_copy.__dict__:
                record_copy.msg = record_copy.__dict__["color_message"]
                record_copy.__dict__["message"] = record_copy.getMessage()
        record_copy.__dict__["levelprefix"] = levelname
        return super().formatMessage(record_copy)


class Bcolors:
    RESET_ALL = '\033[0m'
    BOLD = '\033[1m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'


def configure_default_logging(
    stdout_level=None,
    force=True,
    print_thread_id=False,
    setup_debug_logger=False,
    custom_format_string: str = '',
):
    if not stdout_level:
        stdout_level = logging.INFO

    # make sure to have a clean root logger (in case setup is called multiple times)
    if force:
        handlers = logging.root.handlers
        for h in handlers:
            logging.root.removeHandler(h)
            h.close()

    sh = logging.StreamHandler()
    sh.setLevel(stdout_level)

    if custom_format_string:
        sh.setFormatter(CCFormatter(fmt=custom_format_string))
    else:
        sh.setFormatter(CCFormatter(fmt=default_fmt_string(print_thread_id=print_thread_id)))
    logging.root.addHandler(hdlr=sh)

    if setup_debug_logger:
        dh = setup_debug_handler(
            custom_format_string=custom_format_string,
            print_thread_id=print_thread_id,
        )
        logging.root.addHandler(hdlr=dh)
        logging.root.setLevel(level=logging.DEBUG)
    else:
        logging.root.setLevel(level=stdout_level)

    # both too verbose ...
    logging.getLogger('github3').setLevel(logging.WARNING)
    logging.getLogger('elasticsearch').setLevel(logging.WARNING)


def default_fmt_string(print_thread_id: bool=False):
    ptid = print_thread_id
    return f'%(asctime)s [%(levelprefix)s] {"TID:%(thread)d " if ptid else ""}%(name)s: %(message)s'


def disable_logging(
    log_levels: tuple = (
        logging.CRITICAL,
        logging.ERROR,
        logging.WARNING,
        logging.INFO,
        logging.DEBUG,
    ),
):
    for log_level in log_levels:
        logging.disable(log_level)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import enum
import typing

from concourse.model.step import (
    PipelineStep,
    PrivilegeMode,
    PullRequestNotificationPolicy,
)
from concourse.model.base import (
    AttributeSpec,
    ModelBase,
    ScriptType,
    Trait,
    TraitTransformer,
)
from concourse.model.job import (
    JobVariant,
)
from model.base import ModelValidationError

import concourse.model.traits.component_descriptor
import concourse.model.traits.images

OciImageCfg = concourse.model.traits.images.OciImageCfg


class MergePolicy(enum.Enum):
    MANUAL = 'manual'
    AUTO_MERGE = 'auto_merge'


class MergeMethod(enum.StrEnum):
    MERGE = 'merge'
    REBASE = 'rebase'
    SQUASH = 'squash'


class UpstreamUpdatePolicy(enum.Enum):
    STRICTLY_FOLLOW = 'strictly_follow'
    ACCEPT_HOTFIXES = 'accept_hotfixes'


MERGE_POLICY_CONFIG_ATTRIBUTES = (
    AttributeSpec.optional(
        name='component_names',
        default=[],
        type=typing.List[str],
        doc=(
            'a sequence of regular expressions. This merge policy will be applied to matching '
            'component names. Matches all component names by default'
        )
    ),
    AttributeSpec.optional(
        name='merge_mode',
        default='manual',
        type=MergePolicy,
        doc='whether or not created PRs should be automatically merged',
    ),
    AttributeSpec.optional(
        name='merge_method',
        default='merge',
        type=MergeMethod,
        doc=(
            'The method to use when merging PRs automatically. For more details, check '
            '`the GitHub documentation for merge-methods <https://docs.github.com/en/repositories/configuring-branches-and-merges-in-your-repository/configuring-pull-request-merges/about-merge-methods-on-github>`_.' # noqa:E501
        )
    ),
)


class MergePolicyConfig(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return MERGE_POLICY_CONFIG_ATTRIBUTES

    def component_names(self):
        # handle default here
        # TODO: refactor default arg handling. User-given values should *replace* defaults, not
        #       add to them
        if not self.raw['component_names']:
            return ['.*']

        return self.raw['component_names']

    def merge_mode(self):
        return MergePolicy(self.raw['merge_mode'])

    def merge_method(self):
        return MergeMethod(self.raw['merge_method'])


ATTRIBUTES = (
    AttributeSpec.optional(
        name='set_dependency_version_script',
        default='.ci/set_dependency_version',
        doc='configures the path to set_dependency_version script',
    ),
    AttributeSpec.optional(
        name='set_dependency_version_script_container_image',
        default=None,
        doc='''
            if specified, the `set_dependency_version_script` will be executed in a separate
            OCI container using the specified container image.
            main_repository will be mounted at /mnt/main_repo
        ''',
        type=concourse.model.traits.images.OciImageCfg,
    ),
    AttributeSpec.optional(
        name='upstream_component_name',
        default=None, # defaults to main repository
        doc='configures the upstream component',
    ),
    AttributeSpec.optional(
        name='upstream_update_policy',
        default=UpstreamUpdatePolicy.STRICTLY_FOLLOW,
        doc='configures the upstream component update policy',
    ),
    AttributeSpec.deprecated(
        name='merge_policy',
        default=None,
        doc='whether or not created PRs should be automatically merged. **deprecated**',
        type=MergePolicy,
    ),
    AttributeSpec.optional(
        name='merge_policies',
        default=(),
        doc=(
            'merge policies to apply to detected component upgrades. By default, upgrade '
            'pull-requests must be merged manually'
        ),
        type=typing.List[MergePolicyConfig],
    ),
    AttributeSpec.optional(
        name='after_merge_callback',
        default=None,
        doc='callback to be invoked after auto-merge',
    ),
    AttributeSpec.optional(
        name='ignore_prerelease_versions',
        default=True,
        doc=(
            'ignores prerelease versions like "0.1.0-dev-abc" and '
            'only creates upgrade pr\'s for finalized versions.'
        ),
        type=bool,
    ),
    AttributeSpec.optional(
        name='vars',
        default={},
        doc='env vars to pass to after_merge_callback (similar to step\'s vars)',
        type=dict,
    ),
    AttributeSpec.optional(
        name='pullrequest_body_suffix',
        default=None,
        doc='optional suffix to be appended to created upgrade-pullrequest-bodies',
        type=str,
    ),
)


class UpdateComponentDependenciesTrait(Trait):
    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def set_dependency_version_script_path(self):
        return self.raw['set_dependency_version_script']

    def set_dependency_version_script_container_image(self) -> typing.Optional[OciImageCfg]:
        if raw := self.raw.get('set_dependency_version_script_container_image'):
            return OciImageCfg(raw_dict=raw)
        else:
            return None

    def upstream_component_name(self):
        return self.raw.get('upstream_component_name')

    def upstream_update_policy(self):
        return UpstreamUpdatePolicy(self.raw.get('upstream_update_policy'))

    def merge_policies(self):
        # handle default here
        # TODO: refactor default arg handling. User-given values should *replace* defaults, not
        #       add to them
        if not self.raw.get('merge_policies'):
            if not self.raw.get('merge_policy'):
                return [
                    MergePolicyConfig({
                        'component_names': ['.*'],
                        'merge_mode': 'manual',
                        'merge_method': 'merge',
                    })]

            # preserve legacy behaviour
            # TODO rm
            else:
                return [
                    MergePolicyConfig({
                        'component_names': ['.*'],
                        'merge_mode': self.raw['merge_policy'],
                        'merge_method': 'merge',
                    })
                ]

        else:
            return [
                MergePolicyConfig(cfg) for cfg in self.raw['merge_policies']
            ]

    def after_merge_callback(self):
        return self.raw.get('after_merge_callback')

    def ignore_prerelease_versions(self):
        return self.raw.get('ignore_prerelease_versions')

    def vars(self):
        return self.raw['vars']

    @property
    def pullrequest_body_suffix(self) -> str | None:
        return self.raw.get('pullrequest_body_suffix', None)

    def transformer(self):
        return UpdateComponentDependenciesTraitTransformer(trait=self)

    def validate(self):
        super().validate()
        if self.raw.get('merge_policy') and self.raw.get('merge_policies'):
            raise ModelValidationError(
                "Only one of 'merge_policy' and 'merge_policies' is allowed."
            )
        for config in self.merge_policies():
            config.validate()


class UpdateComponentDependenciesTraitTransformer(TraitTransformer):
    name = 'update_component_deps'

    def __init__(self, trait, *args, **kwargs):
        self.trait = trait
        super().__init__(*args, **kwargs)

    @classmethod
    def order_dependencies(cls):
        return {'component_descriptor'}

    @classmethod
    def dependencies(cls):
        return {'component_descriptor'}

    def inject_steps(self):
        if self.trait.set_dependency_version_script_container_image():
            privilege_mode = PrivilegeMode.PRIVILEGED
        else:
            privilege_mode = PrivilegeMode.UNPRIVILEGED

        # declare no dependencies --> run asap, but do not block other steps
        self.update_component_deps_step = PipelineStep(
                name='update_component_dependencies',
                raw_dict={
                    'privilege_mode': privilege_mode,
                },
                is_synthetic=True,
                pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
                injecting_trait_name=self.name,
                script_type=ScriptType.PYTHON3
        )
        self.update_component_deps_step.add_input(
            name=concourse.model.traits.component_descriptor.DIR_NAME,
            variable_name=concourse.model.traits.component_descriptor.ENV_VAR_NAME,
        )
        self.update_component_deps_step.set_timeout(duration_string='30m')

        for name, value in self.trait.vars().items():
            self.update_component_deps_step.variables()[name] = value

        yield self.update_component_deps_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # our step depends on dependendency descriptor step
        component_descriptor_step = pipeline_args.step(
            concourse.model.traits.component_descriptor.DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME
        )
        self.update_component_deps_step._add_dependency(component_descriptor_step)

        upstream_component_name = self.trait.upstream_component_name()
        if upstream_component_name:
            self.update_component_deps_step.variables()['UPSTREAM_COMPONENT_NAME'] = '"{cn}"'.format(
                cn=upstream_component_name,
            )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import collections
import datetime
import deprecated
import enum
import io
import re
import sys

import typing
from typing import Iterable, Tuple

import requests

import github3
import github3.issues
from github3.exceptions import NotFoundError
from github3.github import GitHub
from github3.orgs import Team
from github3.pulls import PullRequest
from github3.repos.release import Release

import ocm
import ccc.github
import ci.util
import version

from model.github import GithubConfig


class RepoPermission(enum.Enum):
    PULL = "pull"
    PUSH = "push"
    ADMIN = "admin"


class GitHubRepoBranch:
    '''Instances of this class represent a specific branch of a given GitHub repository.
    '''
    def __init__(
        self,
        github_config: GithubConfig,
        repo_owner: str,
        repo_name: str,
        branch: str,
    ):
        self._github_config = ci.util.not_none(github_config)
        self._repo_owner = ci.util.not_empty(repo_owner)
        self._repo_name = ci.util.not_empty(repo_name)
        self._branch = ci.util.not_empty(branch)

    def github_repo_path(self):
        return f'{self._repo_owner}/{self._repo_name}'

    def github_config(self):
        return self._github_config

    def repo_owner(self):
        return self._repo_owner

    def repo_name(self):
        return self._repo_name

    def branch(self):
        return self._branch


class RepositoryHelperBase:
    GITHUB_TIMESTAMP_UTC_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

    def __init__(
        self,
        owner: str,
        name: str,
        default_branch: str='master',
        github_cfg: GithubConfig=None,
        github_api: GitHub=None,
    ):
        '''
        Args:
            owner (str):    repository owner (also called organisation in GitHub)
            name (str):     repository name
            default_branch (str): branch to use for operations when not specified
            github_cfg (GithubConfig): cfg to construct github api object from
            github_api (GitHub): github api to use

        Exactly one of `github_cfg` and `github_api` must be passed as argument.
        Passing a GitHub object is more flexible (but less convenient).
        '''
        if not (bool(github_cfg) ^ bool(github_api)):
            raise ValueError('exactly one of github_api and github_cfg must be given')

        if github_cfg:
            self.github = ccc.github.github_api(github_cfg)
        else:
            self.github = github_api

        self.repository = self._create_repository(
            owner=owner,
            name=name
        )
        self.owner = owner
        self.repository_name = name

        self.default_branch = default_branch

    def _create_repository(self, owner: str, name: str):
        try:
            repository = self.github.repository(
                    owner=owner,
                    repository=name
            )
            return repository
        except NotFoundError as nfe:
            try:
                gh_user = self.github.me().name
            except:
                gh_user = 'failed to determine current user'

            raise RuntimeError(
                f'failed to retrieve repository {owner}/{name} {gh_user=}',
                nfe,
            )


class UpgradePullRequest:
    def __init__(
        self,
        pull_request: PullRequest,
        from_ref: typing.Union[ocm.Resource, ocm.ComponentReference],
        to_ref: typing.Union[ocm.Resource, ocm.ComponentReference],
    ):
        self.pull_request = ci.util.not_none(pull_request)

        if from_ref.name != to_ref.name:
            raise ValueError(f'reference name mismatch {from_ref.name=} {to_ref.name=}')
        if (isinstance(from_ref, ocm.Resource) and isinstance(to_ref, ocm.Resource) and
            from_ref.type != to_ref.type
            ) or \
            type(from_ref) != type(to_ref):
            raise ValueError(f'reference types do not match: {from_ref=} {to_ref=}')

        self.ref_name = from_ref.name

        self.from_ref = from_ref
        self.to_ref = to_ref
        if isinstance(from_ref, ocm.Resource):
            if isinstance(from_ref.type, enum.Enum):
                self.reference_type_name = from_ref.type.value
            elif isinstance(from_ref.type, str):
                self.reference_type_name = from_ref.type
            else:
                raise ValueError(from_ref.type)
        elif isinstance(from_ref, ocm.ComponentReference):
            self.reference_type_name = 'component'
        else:
            raise NotImplementedError(from_ref.type)

    def is_downgrade(self) -> bool:
        from_ver = version.parse_to_semver(self.from_ref.version)
        to_ver = version.parse_to_semver(self.to_ref.version)
        return from_ver > to_ver

    def is_obsolete(
        self,
        reference_component: ocm.Component,
    ):
        '''returns a boolean indicating whether or not this Upgrade PR is "obsolete"

        A Upgrade is considered to be obsolete, iff the following conditions hold true:
        - the reference product contains a component reference with the same name
        - the destination version is greater than the greatest reference component version
        '''
        # find matching component versions
        if not isinstance(reference_component, ocm.Component):
            raise TypeError(reference_component)

        if self.reference_type_name == 'component':
            reference_refs = sorted(
                [
                    rc for rc in reference_component.componentReferences
                    if rc.componentName == self.ref_name
                ],
                key=lambda r: version.parse_to_semver(r.version)
            )

            if not reference_refs:
                return False # special case: we have a new reference

            greatest_reference_version = version.parse_to_semver(reference_refs[-1].version)

        else:
            raise NotImplementedError

        # PR is obsolete if same or newer component version is already configured in reference
        return greatest_reference_version >= version.parse_to_semver(self.to_ref.version)

    def target_matches(
        self,
        reference: typing.Tuple[ocm.ComponentReference, ocm.Resource],
        reference_version: str = None,
    ):
        if not isinstance(reference, ocm.ComponentReference) and not \
                isinstance(reference, ocm.Resource):
            raise TypeError(reference)

        if isinstance(reference, ocm.ComponentReference):
            if self.reference_type_name != 'component':
                return False
            if reference.componentName != self.ref_name:
                return False
        else: # ocm.Resource, already checked above
            if reference.name != self.ref_name:
                return False
            if isinstance(reference.type, enum.Enum):
                reference_type = reference.type.value
            elif isinstance(reference.type, str):
                reference_type = reference.type
            else:
                raise ValueError(reference.type)
            if reference_type != self.reference_type_name:
                return False

        reference_version = reference_version or reference.version
        if reference_version != self.to_ref.version:
            return False

        return True

    def purge(self):
        self.pull_request.close()
        head_ref = 'heads/' + self.pull_request.head.ref
        self.pull_request.repository.ref(head_ref).delete()


def iter_obsolete_upgrade_pull_requests(
    upgrade_pull_requests: typing.Iterable[UpgradePullRequest],
    keep_hotfix_versions: bool=True,
) -> typing.Generator[UpgradePullRequest, None, None]:
    grouped_upgrade_pull_requests = collections.defaultdict(list)

    def group_name(upgrade_pull_request: UpgradePullRequest):
        '''
        calculate groupname, depending on whether or not we should keep hotfix_versions;
        for each upgrade-pr-group, we keep only exactly one version (the greatest tgt-version);
        therefore, to prevent hotfix-upgrades from being removed, collect hotfixes in a separate
        group.
        '''
        cname = upgrade_pull_request.to_ref.componentName

        if not keep_hotfix_versions:
            return cname

        from_version = version.parse_to_semver(upgrade_pull_request.from_ref.version)
        to_version = version.parse_to_semver(upgrade_pull_request.to_ref.version)

        if from_version.major != to_version.major:
            return cname # not a hotfix
        if from_version.minor != to_version.minor:
            return cname # not a hotfix (hardcode hotfixes differ at patchlevel, always)

        # we have a hotfix version (patchlevel differs)
        return f'{cname}:{from_version.major}.{from_version.minor}'

    for upgrade_pull_request in upgrade_pull_requests:
        if upgrade_pull_request.pull_request.state != 'open':
            continue
        name = group_name(upgrade_pull_request)
        grouped_upgrade_pull_requests[name].append(upgrade_pull_request)

    for upgrade_pull_request_group in grouped_upgrade_pull_requests.values():
        if len(upgrade_pull_request_group) < 2:
            continue

        # greatest version will be sorted as last element
        ordered_by_version = sorted(
            upgrade_pull_request_group,
            key=lambda upr: version.parse_to_semver(upr.to_ref.version),
        )

        greatest_version = version.parse_to_semver(ordered_by_version[-1].to_ref.version)
        for upgrade_pr in ordered_by_version:
            if version.parse_to_semver(upgrade_pr.to_ref.version) < greatest_version:
                yield upgrade_pr


class PullRequestUtil(RepositoryHelperBase):
    _pr_title_pattern = re.compile(r'^\[ci:(\S*):(\S*):(\S*)->(\S*)\]$')

    @staticmethod
    def calculate_pr_title(
            reference: ocm.ComponentReference,
            from_version: str,
            to_version: str,
    ) -> str:
        if not isinstance(reference, ocm.ComponentReference):
            raise TypeError(reference)

        type_name = 'component'
        reference_name = reference.componentName

        return f'[ci:{type_name}:{reference_name}:{from_version}->{to_version}]'

    def _pr_to_upgrade_pull_request(
        self,
        pull_request,
        pattern: re.Pattern=None,
    ) -> UpgradePullRequest:
        ci.util.not_none(pull_request)

        if not pattern:
            pattern = self._pr_title_pattern

        match = pattern.fullmatch(pull_request.title)
        if match is None:
            raise ValueError("PR-title '{t}' did not match title-schema".format(
                t=pull_request.title)
            )

        reference_type_name = match.group(1)
        if not reference_type_name:
            # backwards compatibility hack
            reference_type_name = 'component'

        if not reference_type_name == 'component':
            raise NotImplementedError(reference_type_name) # todo: support all resources

        ref_name = match.group(2)
        from_version = match.group(3)
        to_version = match.group(4)

        from_ref = ocm.ComponentReference(
            name=ref_name,
            componentName=ref_name,
            version=from_version,
        )
        to_ref = ocm.ComponentReference(
            name=ref_name,
            componentName=ref_name,
            version=to_version,
        )

        return UpgradePullRequest(
            pull_request=pull_request,
            from_ref=from_ref,
            to_ref=to_ref,
        )

    def enumerate_upgrade_pull_requests(
        self,
        state: str='all',
        pattern: re.Pattern=None,
    ):
        '''returns a sequence of `UpgradePullRequest` for all found pull-requests
        '''
        def has_upgrade_pr_title(pull_request):
            return bool(pattern.fullmatch(pull_request.title))

        def pr_to_upgrade_pr(pull_request):
            return self._pr_to_upgrade_pull_request(
                pull_request=pull_request,
                pattern=pattern,
            )

        def strip_title(pull_request):
            pull_request.title = pull_request.title.strip()
            return pull_request

        if not pattern:
            pattern = self._pr_title_pattern

        parsed_prs = ci.util.FluentIterable(
            self.repository.pull_requests(
                state=state,
                number=128, # avoid issueing more than one github-api-request
            )
        ) \
            .map(strip_title) \
            .filter(has_upgrade_pr_title) \
            .map(pr_to_upgrade_pr) \
            .filter(lambda e: e) \
            .as_list()
        return parsed_prs

    def retrieve_pr_template_text(self):
        '''Return the content for the PR template file looking in predefined directories.
        If no template is found None is returned.
        '''
        pattern = re.compile(r"(pull_request_template)(\..{1,3})?$")
        directories = ['.github', '.', 'docs']
        for directory in directories:
            try:
                for filename, content in self.repository.directory_contents(directory):
                    if pattern.match(filename):
                        content.refresh()
                        return content.decoded.decode('utf-8')
            except github3.exceptions.NotFoundError:
                pass  # directory does not exist

        return None


class GitHubRepositoryHelper(RepositoryHelperBase):
    def create_or_update_file(
        self,
        file_path: str,
        file_contents: str,
        commit_message: str,
        branch: str=None,
    ) -> str:
        if branch is None:
            branch = self.default_branch

        try:
            contents = self.retrieve_file_contents(file_path=file_path, branch=branch)
        except NotFoundError:
            contents = None # file did not yet exist

        if contents:
            decoded_contents = contents.decoded.decode('utf-8')
            if decoded_contents == file_contents:
                # Nothing to do
                return ci.util.info(
                    'Repository file contents are identical to passed file contents.'
                )
            else:
                response = contents.update(
                    message=commit_message,
                    content=file_contents.encode('utf-8'),
                    branch=branch,
                )
        else:
            response = self.repository.create_file(
                path=file_path,
                message=commit_message,
                content=file_contents.encode('utf-8'),
                branch=branch,
            )
        return response['commit'].sha

    @staticmethod
    def from_githubrepobranch(
        githubrepobranch: GitHubRepoBranch,
    ):
        return GitHubRepositoryHelper(
            github_cfg=githubrepobranch.github_config(),
            owner=githubrepobranch.repo_owner(),
            name=githubrepobranch.repo_name(),
            default_branch=githubrepobranch.branch(),
        )

    def retrieve_file_contents(self, file_path: str, branch: str=None):
        if branch is None:
            branch = self.default_branch

        return self.repository.file_contents(
            path=file_path,
            ref=branch,
        )

    def retrieve_text_file_contents(
        self,
        file_path: str,
        branch: str=None,
        encoding: str='utf-8',
    ):
        if branch is None:
            branch = self.default_branch

        contents = self.retrieve_file_contents(file_path, branch)
        return contents.decoded.decode(encoding)

    def create_tag(
        self,
        tag_name: str,
        tag_message: str,
        repository_reference: str,
        author_name: str,
        author_email: str,
        repository_reference_type: str='commit'
    ):
        author = {
            'name': author_name,
            'email': author_email,
            'date': datetime.datetime.now(datetime.timezone.utc)
                    .strftime(self.GITHUB_TIMESTAMP_UTC_FORMAT)
        }
        self.repository.create_tag(
            tag=tag_name,
            message=tag_message,
            sha=repository_reference,
            obj_type=repository_reference_type,
            tagger=author
        )

    MAXIMUM_GITHUB_RELEASE_BODY_LENGTH = 25000
    '''The limit is not documented explicitly in the GitHub docs. To see it, the error returned by
    GitHub when creating a release with more then the allowed number of characters must be
    looked at.'''

    def _replacement_release_notes(
        self,
        asset_url: str,
        component_name: str,
        component_version: str,
    ):
        return (
            f'The release-notes for component **{component_name}** in version '
            f'**{component_version}** exceeded the maximum length of '
            f'{self.MAXIMUM_GITHUB_RELEASE_BODY_LENGTH} characters allowed by GitHub for '
            'release-bodies.\n'
            f'They have been uploaded as release-asset and can be found at {asset_url}.'
        )

    RELEASE_NOTES_ASSET_NAME = 'release_notes.md'

    def create_release(
        self,
        tag_name: str,
        body: str,
        draft: bool=False,
        prerelease: bool=False,
        name: str=None,
        component_name: str=None,
        component_version: str=None,
    ):
        if len(body) < self.MAXIMUM_GITHUB_RELEASE_BODY_LENGTH:
            return self.repository.create_release(
                tag_name=tag_name,
                body=body,
                draft=draft,
                prerelease=prerelease,
                name=name
            )
        else:
            # release notes are too large to be added to the github-release directly. As a work-
            # around, attach them to the release and write an appropriate release-body pointing
            # towards the attached asset.
            # For draft releases, the url cannot be calculated easily beforehand, so create the
            # release, attach the notes and then retrieve the URL via the asset-object.
            release = self.repository.create_release(
                tag_name=tag_name,
                body='',
                draft=draft,
                prerelease=prerelease,
                name=name
            )
            release_notes_asset = release.upload_asset(
                content_type='text/markdown',
                name=self.RELEASE_NOTES_ASSET_NAME,
                asset=body.encode('utf-8'),
                label='Release Notes',
            )
            release.edit(
                body=self._replacement_release_notes(
                    asset_url=release_notes_asset.browser_download_url,
                    component_name=component_name,
                    # Fallback: use tag_name if version not explicitly given
                    component_version=component_version or tag_name,
                )
            )
            return release

    def delete_releases(
        self,
        release_names: typing.Iterable[str],
    ):
        for release in self.repository.releases():
            if release.name in release_names:
                release.delete()

    def create_draft_release(
        self,
        name: str,
        body: str,
        component_name: str=None,
        component_version: str=None,
    ):
        return self.create_release(
            tag_name='',
            name=name,
            body=body,
            draft=True,
            component_name=component_name,
            component_version=component_version,
        )

    def promote_draft_release(
        self,
        draft_release,
        release_tag,
        release_version,
        component_name: str=None,
    ):
        draft_release.edit(
            tag_name=release_tag,
            body=None,
            draft=False,
            prerelease=False,
            name=release_version,
        )

        # If there is a release-notes asset attached, we need to update the release-notes after
        # promoting the release so that the contained URL is adjusted as well
        release_notes_asset = next(
            (a for a in draft_release.assets() if a.name == self.RELEASE_NOTES_ASSET_NAME),
            None,
        )
        if release_notes_asset:
            draft_release.edit(
                body=self._replacement_release_notes(
                    asset_url=release_notes_asset.browser_download_url,
                    component_name=component_name,
                    component_version=release_version,
                )
            )

    def update_release_notes(
        self,
        tag_name: str,
        component_name: str,
        body: str,
    ) -> bool:
        ci.util.not_empty(tag_name)

        release = self.repository.release_from_tag(tag_name)
        if not release:
            raise RuntimeError(
                f"No release with tag '{tag_name}' found "
                f"in repository {self.repository}"
            )

        if len(body) < self.MAXIMUM_GITHUB_RELEASE_BODY_LENGTH:
            release.edit(body=body)
        else:
            release_notes_asset = next(
                (a for a in release.assets() if a.name == self.RELEASE_NOTES_ASSET_NAME),
                None,
            )
            # Clean up any attached release-note-asset
            if release_notes_asset:
                release_notes_asset.delete()

            release_notes_asset = release.upload_asset(
                content_type='text/markdown',
                name=self.RELEASE_NOTES_ASSET_NAME,
                asset=body.encode('utf-8'),
                label='Release Notes',
            )
            release.edit(
                body=self._replacement_release_notes(
                    asset_url=release_notes_asset.browser_download_url,
                    component_version=tag_name,
                    component_name=component_name)
            )

        return release

    def draft_release_with_name(
        self,
        name: str
    ) -> Release:
        # if there are more than 1021 releases, github(.com) will return http-500 one requesting
        # additional releases. As this limit is typically not reached, hardcode limit for now
        # in _most_ cases, most recent releases are returned first, so this should hardly ever
        # be an actual issue
        max_releases = 1020
        for release in self.repository.releases(number=max_releases):
            if not release.draft:
                continue
            if release.name == name:
                return release

    def tag_exists(
        self,
        tag_name: str,
    ):
        ci.util.not_empty(tag_name)
        try:
            self.repository.ref('tags/' + tag_name)
            return True
        except NotFoundError:
            return False

    def retrieve_asset_contents(self, release_tag: str, asset_label: str):
        ci.util.not_none(release_tag)
        ci.util.not_none(asset_label)

        release = self.repository.release_from_tag(release_tag)
        for asset in release.assets():
            if asset.label == asset_label or asset.name == asset_label:
                break
        else:
            response = requests.Response()
            response.status_code = 404
            response.json = lambda: {'message':'no asset with label {} found'.format(asset_label)}
            raise NotFoundError(resp=response)

        buffer = io.BytesIO()
        asset.download(buffer)
        return buffer.getvalue().decode()

    def release_versions(self):
        for tag_name in self._release_tags():
            try:
                version.parse_to_semver(tag_name)
                yield tag_name
                # XXX should rather return a "Version" object, containing both parsed and original
            except ValueError:
                pass # ignore

    def _release_tags(self):
        for release in self.repository.releases():
            if release.draft or release.prerelease:
                continue
            if not (tag := release.tag_name):
                continue

            yield tag

    def search_issues_in_repo(self, query: str):
        query = f'repo:{self.owner}/{self.repository_name} {query}'
        search_result = self.github.search_issues(query)
        return search_result

    def is_pr_created_by_org_member(self, pull_request_number):
        pull_request = self.repository.pull_request(pull_request_number)
        user_login = pull_request.user.login
        return self.is_org_member(self.owner, user_login)

    def add_labels_to_pull_request(self, pull_request_number, *labels):
        pull_request = self.repository.pull_request(pull_request_number)
        pull_request.issue().add_labels(*labels)

    def remove_label_from_pull_request(self, pull_request_number, label):
        pull_request = self.repository.pull_request(pull_request_number)
        pull_request.issue().remove_label(label)

    def add_comment_to_pr(self, pull_request_number, comment):
        pull_request = self.repository.pull_request(pull_request_number)
        pull_request.create_comment(comment)

    def is_org_member(self, organization_name, user_login):
        organization = self.github.organization(organization_name)
        return organization.is_member(user_login)

    def is_team_member(self, team_name, user_login) -> bool:
        '''Returns a bool indicating team-membership to the given team for the given user-login

        The team-name is expected in the format `<org-name>/<team-name>`.
        Note: If the team cannot be seen by the user used for the lookup or does not exist `False`
        will be returned.
        '''
        o, t = team_name.split('/')
        org = self.github.organization(o)
        try:
            team = org.team_by_name(t)
            team.membership_for(user_login)
        except github3.exceptions.NotFoundError:
            return False
        else:
            return True

    def delete_outdated_draft_releases(self) -> Iterable[Tuple[github3.repos.release.Release, bool]]:
        '''Find outdated draft releases and try to delete them

        Yields tuples containing a release and a boolean indicating whether its deletion was
        successful.

        A draft release is considered outdated iff:
        1: its version is smaller than the greatest release version (according to semver) AND
            2a: it is NOT a hotfix draft release AND
            2b: there are no hotfix draft releases with the same major and minor version
            OR
            3a: it is a hotfix draft release AND
            3b: there is a hotfix draft release of greater version (according to semver)
                with the same major and minor version
        '''

        releases = [release for release in self.repository.releases(number=20)]
        non_draft_releases = [release for release in releases if not release.draft]
        draft_releases = [release for release in releases if release.draft]
        greatest_release_version = find_greatest_github_release_version(non_draft_releases)

        if greatest_release_version is not None:
            draft_releases_to_delete = outdated_draft_releases(
                    draft_releases=draft_releases,
                    greatest_release_version=greatest_release_version,
            )
        else:
            draft_releases_to_delete = []

        for release in draft_releases_to_delete:
            yield release, release.delete()


@deprecated.deprecated
def github_cfg_for_hostname(cfg_factory, host_name, require_labels=('ci',)): # XXX unhardcode label
    return ccc.github.github_cfg_for_hostname(
        host_name=host_name,
        cfg_factory=cfg_factory,
        require_labels=require_labels,
    )


@deprecated.deprecated
def _create_github_api_object(github_cfg):
    return ccc.github.github_api(github_cfg=github_cfg)


def branches(
    github_cfg,
    repo_owner: str,
    repo_name: str,
):
    github_api = ccc.github.github_api(github_cfg=github_cfg)
    repo = github_api.repository(repo_owner, repo_name)
    return list(map(lambda r: r.name, repo.branches()))


def retrieve_email_addresses(
    github_cfg: GithubConfig,
    github_users: typing.Sequence[str] | typing.Collection[str],
    out_file: str=None
):
    github = ccc.github.github_api(github_cfg=github_cfg)

    def retrieve_email(username: str):
        user = github.user(username)
        return user.email

    fh = open(out_file, 'w') if out_file else sys.stdout

    email_addresses_count = 0

    for email_address in filter(None, map(retrieve_email, github_users)):
        fh.write(email_address + '\n')
        email_addresses_count += 1

    ci.util.verbose('retrieved {sc} email address(es) from {uc} user(s)'.format(
        sc=email_addresses_count,
        uc=len(github_users)
    )
    )


def _retrieve_team_by_name_or_none(
    organization: github3.orgs.Organization,
    team_name: str
) -> Team:

    team_list = list(filter(lambda t: t.name == team_name, organization.teams()))
    return team_list[0] if team_list else None


def find_greatest_github_release_version(
    releases: list[Release],
    warn_for_unparseable_releases: bool = True,
    ignore_prerelease_versions: bool = False,
):
    # currently, non-draft-releases are not created with a name by us. Use the tag name as fallback
    release_versions = [
        release.name if release.name else release.tag_name
        for release in releases
    ]

    def filter_non_semver_parseable_releases(release_name):
        try:
            version.parse_to_semver(release_name)
            return True
        except ValueError:
            if warn_for_unparseable_releases:
                ci.util.warning(f'ignoring release {release_name=} (not semver)')
            return False

    release_versions = [
        name for name in filter(filter_non_semver_parseable_releases, release_versions)
    ]

    release_version_infos = [
        version.parse_to_semver(release_version)
        for release_version in release_versions
    ]
    latest_version = version.find_latest_version(
        versions=release_version_infos,
        ignore_prerelease_versions=ignore_prerelease_versions,
    )
    if latest_version:
        return str(latest_version)
    else:
        return None


def outdated_draft_releases(
    draft_releases: list[Release],
    greatest_release_version: str,
):
    '''Find outdated draft releases from a list of draft releases and return them. This is achieved
    by partitioning the release versions according to their joined major and minor version.
    Partitions are then checked:
        - if there is only a single release in a partition it is either a hotfix release
            (keep corresponding release) or it is not (delete if it is not the greatest release
            according to semver)
        - if there are multiple releases versions in a partition, keep only the release
            corresponding to greatest (according to semver)
    '''

    greatest_release_version_info = version.parse_to_semver(greatest_release_version)

    def _has_semver_draft_prerelease_label(release_name):
        version_info = version.parse_to_semver(release_name)
        if version_info.prerelease != 'draft':
            return False
        return True

    autogenerated_draft_releases = [
        release for release in draft_releases
        if release.name
        and version.is_semver_parseable(release.name)
        and _has_semver_draft_prerelease_label(release.name)
    ]

    draft_release_version_infos = [
        version.parse_to_semver(release.name)
        for release in autogenerated_draft_releases
    ]

    def _yield_outdated_version_infos_from_partition(partition):
        if len(partition) == 1:
            version_info = partition.pop()
            if version_info < greatest_release_version_info and version_info.patch == 0:
                yield version_info
        else:
            yield from [
                version_info
                for version_info in partition[1:]
            ]

    outdated_version_infos = list()
    for partition in version.partition_by_major_and_minor(draft_release_version_infos):
        outdated_version_infos.extend(_yield_outdated_version_infos_from_partition(partition))

    outdated_draft_releases = [
        release
        for release in autogenerated_draft_releases
        if version.parse_to_semver(release.name) in outdated_version_infos
    ]

    return outdated_draft_releases


def close_issue(
    issue: github3.issues.ShortIssue,
) -> bool:
    '''
    handle known corner-cases where regular issue close will fail
    comment on issue if closing still fails

    returns True if close was successful, False otherwise
    '''
    def try_close() -> bool:
        try:
            return issue.edit(
                state='closed',
            )

        except github3.exceptions.UnprocessableEntity:
            # likely that assignee was suspended from github

            if not issue.assignees:
                raise

            issue.remove_assignees(issue.assignees)
            return issue.edit(
                state='closed',
            )

    closed = try_close()
    if not closed:
        issue.create_comment('unable to close ticket')

    return closed



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import ctt.process_dependencies as process_dependencies


def test_processor_instantiation(tmpdir):
    tmpfile = tmpdir.join('a_file')
    tmpfile.write('')  # touch

    cfg = {
        'filter': {
            'type': 'ImageFilter',
            'kwargs': {
                'include_image_refs': ['^aaa'],
            },
        },
        'processor': {
            'type': 'FileFilter',
            'kwargs': {
                'filter_files': [tmpfile],
            },
        },
        'upload': {
            'type': 'PrefixUploader',
            'kwargs': {
                'prefix': 'a/prefix',
            },
        },
    }

    _ = process_dependencies.processing_pipeline(cfg)

    # test shared processor
    shared_proc = {'shared_p': cfg['processor']}
    cfg['processor'] = 'shared_p'

    _ = process_dependencies.processing_pipeline(cfg, shared_processors=shared_proc)

    # revert
    cfg['processor'] = shared_proc['shared_p']

    # test shared uploader
    shared_upld = {'shared_u': cfg['upload']}
    cfg['upload'] = 'shared_u'

    _ = process_dependencies.processing_pipeline(cfg, shared_uploaders=shared_upld)
